"""Command line functions for interacting with MFA models"""
from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING, Optional

from montreal_forced_aligner.config import get_temporary_directory
from montreal_forced_aligner.data import PhoneSetType
from montreal_forced_aligner.exceptions import (
    FileArgumentNotFoundError,
    ModelLoadError,
    ModelTypeNotSupportedError,
    MultipleModelTypesFoundError,
    PretrainedModelNotFoundError,
    RemoteModelNotFoundError,
)
from montreal_forced_aligner.models import MODEL_TYPES, Archive, ModelManager, guess_model_type
from montreal_forced_aligner.utils import configure_logger

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = [
    "inspect_model",
    "validate_args",
    "save_model",
    "run_model",
]


def inspect_model(path: str) -> None:
    """
    Inspect a model and print out metadata information about it

    Parameters
    ----------
    path: str
        Path to model
    """
    working_dir = os.path.join(get_temporary_directory(), "models", "inspect")
    ext = os.path.splitext(path)[1]
    model = None
    if ext == Archive.extensions[0]:  # Figure out what kind of model it is
        a = Archive(path, working_dir)
        model = a.get_subclass_object()
    else:
        for model_class in MODEL_TYPES.values():
            if model_class.valid_extension(path):
                if model_class == MODEL_TYPES["dictionary"]:
                    model = model_class(path, working_dir, phone_set_type=PhoneSetType.AUTO)
                else:
                    model = model_class(path, working_dir)
    if not model:
        raise ModelLoadError(path)
    model.pretty_print()


def save_model(path: str, model_type: str, output_name: Optional[str]) -> None:
    """
    Save a model to pretrained folder for later use

    Parameters
    ----------
    path: str
        Path to model
    model_type: str
        Type of model
    """
    logger = configure_logger("save_model")
    model_name = os.path.splitext(os.path.basename(path))[0]
    model_class = MODEL_TYPES[model_type]
    if output_name:
        out_path = model_class.get_pretrained_path(output_name, enforce_existence=False)
    else:
        out_path = model_class.get_pretrained_path(model_name, enforce_existence=False)
    shutil.copyfile(path, out_path)
    logger.info(
        f"Saved model to f{output_name}, you can now use {output_name} in place of paths in mfa commands."
    )


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    :class:`~montreal_forced_aligner.exceptions.ModelTypeNotSupportedError`
        If the type of model is not supported
    :class:`~montreal_forced_aligner.exceptions.FileArgumentNotFoundError`
        If the file specified is not found
    :class:`~montreal_forced_aligner.exceptions.PretrainedModelNotFoundError`
        If the pretrained model specified is not found
    :class:`~montreal_forced_aligner.exceptions.ModelExtensionError`
        If the extension is not valid for the specified model type
    :class:`~montreal_forced_aligner.exceptions.MultipleModelTypesFoundError`
        If multiple model types match the name
    """
    if args.action == "download":
        if args.model_type not in MODEL_TYPES:
            raise ModelTypeNotSupportedError(args.model_type, MODEL_TYPES)
        elif args.model_type:
            args.model_type = args.model_type.lower()
        if args.name:
            manager = ModelManager()
            manager.refresh_remote()
            available_languages = manager.remote_models[args.model_type]
            if args.name not in available_languages:
                raise RemoteModelNotFoundError(
                    args.name, args.model_type, list(available_languages.keys())
                )
    elif args.action == "list":
        if args.model_type and args.model_type.lower() not in MODEL_TYPES:
            raise ModelTypeNotSupportedError(args.model_type, MODEL_TYPES)
        elif args.model_type:
            args.model_type = args.model_type.lower()
    elif args.action == "inspect":
        if args.model_type and args.model_type not in MODEL_TYPES:
            raise ModelTypeNotSupportedError(args.model_type, MODEL_TYPES)
        elif args.model_type:
            args.model_type = args.model_type.lower()
        possible_model_types = guess_model_type(args.name)
        if not possible_model_types:
            if args.model_type:
                model_class = MODEL_TYPES[args.model_type]
                path = model_class.get_pretrained_path(args.name)
                if path is None:
                    raise PretrainedModelNotFoundError(
                        args.name, args.model_type, model_class.get_available_models()
                    )
            else:
                found_model_types = []
                path = None
                for model_type, model_class in MODEL_TYPES.items():
                    p = model_class.get_pretrained_path(args.name)
                    if p is not None:
                        path = p
                        found_model_types.append(model_type)
                if len(found_model_types) > 1:
                    raise MultipleModelTypesFoundError(args.name, found_model_types)
                if path is None:
                    raise PretrainedModelNotFoundError(args.name)
            args.name = path
        else:
            if not os.path.exists(args.name):
                raise FileArgumentNotFoundError(args.name)
    elif args.action == "save":
        if not os.path.exists(args.path):
            raise FileArgumentNotFoundError(args.path)


def run_model(args: Namespace) -> None:
    """
    Wrapper function for running model utility commands

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    """
    validate_args(args)
    manager = ModelManager(token=getattr(args, "github_token", None))
    if args.action == "download" and args.name:
        manager.download_model(args.model_type, args.name, args.ignore_cache)
    elif args.action == "download":
        manager.print_remote_models(args.model_type)
    elif args.action == "list":
        manager.print_local_models(args.model_type)
    elif args.action == "inspect":
        inspect_model(args.name)
    elif args.action == "save":
        save_model(args.path, args.model_type, args.name)



"""Class definitions for base aligner"""
from __future__ import annotations

import collections
import csv
import functools
import multiprocessing as mp
import os
import time
from queue import Empty
from typing import Dict, List, Optional

import sqlalchemy
import tqdm
from sqlalchemy.orm import joinedload, selectinload, subqueryload

from montreal_forced_aligner.abc import FileExporterMixin
from montreal_forced_aligner.alignment.mixins import AlignMixin
from montreal_forced_aligner.alignment.multiprocessing import (
    AlignmentExtractionArguments,
    AlignmentExtractionFunction,
    ExportTextGridArguments,
    ExportTextGridProcessWorker,
    GeneratePronunciationsArguments,
    GeneratePronunciationsFunction,
    construct_output_path,
)
from montreal_forced_aligner.corpus.acoustic_corpus import AcousticCorpusPronunciationMixin
from montreal_forced_aligner.data import CtmInterval, PronunciationProbabilityCounter, TextFileType
from montreal_forced_aligner.db import (
    Corpus,
    DictBundle,
    Dictionary,
    File,
    PhoneInterval,
    Pronunciation,
    SoundFile,
    Speaker,
    TextFile,
    Utterance,
    Word,
    WordInterval,
    WordType,
)
from montreal_forced_aligner.exceptions import AlignmentExportError
from montreal_forced_aligner.helper import align_phones
from montreal_forced_aligner.textgrid import export_textgrid, output_textgrid_writing_errors
from montreal_forced_aligner.utils import Counter, KaldiProcessWorker, Stopped

__all__ = ["CorpusAligner"]


class CorpusAligner(AcousticCorpusPronunciationMixin, AlignMixin, FileExporterMixin):
    """
    Mixin class that aligns corpora with pronunciation dictionaries

    See Also
    --------
    :class:`~montreal_forced_aligner.corpus.acoustic_corpus.AcousticCorpusPronunciationMixin`
        For dictionary and corpus parsing parameters
    :class:`~montreal_forced_aligner.alignment.mixins.AlignMixin`
        For alignment parameters
    :class:`~montreal_forced_aligner.abc.FileExporterMixin`
        For file exporting parameters
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.export_output_directory = None

    def alignment_extraction_arguments(self) -> List[AlignmentExtractionArguments]:
        """
        Generate Job arguments for
        :class:`~montreal_forced_aligner.alignment.multiprocessing.AlignmentExtractionFunction`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.alignment.multiprocessing.AlignmentExtractionArguments`]
            Arguments for processing
        """
        arguments = []
        for j in self.jobs:
            if not j.has_data:
                continue
            arguments.append(
                AlignmentExtractionArguments(
                    j.name,
                    getattr(self, "db_path", ""),
                    os.path.join(self.working_log_directory, f"get_phone_ctm.{j.name}.log"),
                    self.alignment_model_path,
                    round(self.frame_shift / 1000, 4),
                    self.cleanup_textgrids,
                    j.construct_path_dictionary(self.working_directory, "ali", "ark"),
                    j.construct_path_dictionary(self.data_directory, "text", "int.scp"),
                )
            )

        return arguments

    def export_textgrid_arguments(
        self, output_format: str, include_original_text: bool = False
    ) -> List[ExportTextGridArguments]:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.alignment.multiprocessing.ExportTextGridProcessWorker`

        Parameters
        ----------
        output_format: str, optional
            Format to save alignments, one of 'long_textgrids' (the default), 'short_textgrids', or 'json', passed to praatio

        Returns
        -------
        list[:class:`~montreal_forced_aligner.alignment.multiprocessing.ExportTextGridArguments`]
            Arguments for processing
        """
        return [
            ExportTextGridArguments(
                j.name,
                getattr(self, "db_path", ""),
                os.path.join(self.working_log_directory, f"export_textgrids.{j.name}.log"),
                self.frame_shift,
                self.export_output_directory,
                output_format,
                include_original_text,
            )
            for j in self.jobs
            if j.has_data
        ]

    def generate_pronunciations_arguments(
        self,
    ) -> List[GeneratePronunciationsArguments]:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.alignment.multiprocessing.GeneratePronunciationsFunction`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.alignment.multiprocessing.GeneratePronunciationsArguments`]
            Arguments for processing
        """

        return [
            GeneratePronunciationsArguments(
                j.name,
                getattr(self, "db_path", ""),
                os.path.join(self.working_log_directory, f"generate_pronunciations.{j.name}.log"),
                j.construct_path_dictionary(self.data_directory, "text", "int.scp"),
                j.construct_path_dictionary(self.working_directory, "ali", "ark"),
                self.model_path,
                False,
            )
            for j in self.jobs
            if j.has_data
        ]

    def compute_pronunciation_probabilities(self, compute_silence_probabilities=True):
        """
        Multiprocessing function that computes pronunciation probabilities from alignments

        Parameters
        ----------
        compute_silence_probabilities: bool
            Flag for whether to compute silence probabilities for pronunciations, defaults to True
        See Also
        --------
        :class:`~montreal_forced_aligner.alignment.multiprocessing.GeneratePronunciationsFunction`
            Multiprocessing helper function for each job
        :meth:`.CorpusAligner.generate_pronunciations_arguments`
            Job method for generating arguments for the helper function
        :kaldi_steps:`align_si`
            Reference Kaldi script
        :kaldi_steps:`align_fmllr`
            Reference Kaldi script
        """

        def format_probability(probability_value: float) -> float:
            """Format a probability to have two decimal places and be between 0.01 and 0.99"""
            return min(max(round(probability_value, 2), 0.01), 0.99)

        def format_correction(correction_value: float) -> float:
            """Format a probability correction value to have two decimal places and be  greater than 0.01"""
            correction_value = round(correction_value, 2)
            if correction_value == 0:
                correction_value = 0.01
            return correction_value

        begin = time.time()
        dictionary_counters = {
            dict_id: PronunciationProbabilityCounter()
            for dict_id in self.dictionary_lookup.values()
        }
        self.log_info("Generating pronunciations...")
        arguments = self.generate_pronunciations_arguments()
        with tqdm.tqdm(
            total=self.num_current_utterances, disable=getattr(self, "quiet", False)
        ) as pbar:
            if self.use_mp:
                error_dict = {}
                return_queue = mp.Queue()
                stopped = Stopped()
                procs = []
                for i, args in enumerate(arguments):
                    function = GeneratePronunciationsFunction(args)
                    p = KaldiProcessWorker(i, return_queue, function, stopped)
                    procs.append(p)
                    p.start()
                while True:
                    try:
                        result = return_queue.get(timeout=1)
                        if isinstance(result, Exception):
                            error_dict[getattr(result, "job_name", 0)] = result
                            continue
                        if stopped.stop_check():
                            continue
                    except Empty:
                        for proc in procs:
                            if not proc.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    dict_id, utterance_counter = result
                    dictionary_counters[dict_id].add_counts(utterance_counter)
                    pbar.update(1)
                for p in procs:
                    p.join()
                if error_dict:
                    for v in error_dict.values():
                        raise v
            else:
                self.log_debug("Not using multiprocessing...")
                for args in arguments:
                    function = GeneratePronunciationsFunction(args)
                    for dict_id, utterance_counter in function.run():
                        dictionary_counters[dict_id].add_counts(utterance_counter)
                        pbar.update(1)

        initial_key = ("<s>", "")
        final_key = ("</s>", "")
        lambda_2 = 2
        silence_prob_sum = 0
        initial_silence_prob_sum = 0
        final_silence_correction_sum = 0
        final_non_silence_correction_sum = 0
        with open(
            os.path.join(self.working_log_directory, "pronunciation_probability_calculation.log"),
            "w",
            encoding="utf8",
        ) as log_file, self.session() as session:
            dictionaries = session.query(Dictionary.id)
            dictionary_mappings = []
            for (d_id,) in dictionaries:
                counter = dictionary_counters[d_id]
                log_file.write(f"For {d_id}:\n")
                words = (
                    session.query(Word.word)
                    .filter(Word.dictionary_id == d_id)
                    .filter(Word.word_type != WordType.silence)
                )
                pronunciations = (
                    session.query(Word.word, Pronunciation.pronunciation, Pronunciation.id)
                    .join(Pronunciation.word)
                    .filter(Word.dictionary_id == d_id)
                    .filter(Word.word_type != WordType.silence)
                )
                pron_mapping = {}
                for w, p, p_id in pronunciations:
                    pron_mapping[(w, p)] = {"id": p_id}
                    if w in {initial_key[0], final_key[0], self.silence_word}:
                        continue
                    counter.word_pronunciation_counts[w][p] += 1  # Add one smoothing
                for (w,) in words:
                    if w in {initial_key[0], final_key[0], self.silence_word}:
                        continue
                    pron_counts = counter.word_pronunciation_counts[w]
                    max_value = max(pron_counts.values())
                    for p, c in pron_counts.items():
                        pron_mapping[(w, p)]["probability"] = format_probability(c / max_value)

                if not compute_silence_probabilities:
                    log_file.write("Skipping silence calculations")
                    continue
                silence_count = sum(counter.silence_before_counts.values())
                non_silence_count = sum(counter.non_silence_before_counts.values())
                log_file.write(f"Total silence count was {silence_count}\n")
                log_file.write(f"Total non silence count was {non_silence_count}\n")
                silence_probability = silence_count / (silence_count + non_silence_count)
                silence_prob_sum += silence_probability
                silence_probabilities = {}
                for w, p, _ in pronunciations:
                    count = counter.silence_following_counts[(w, p)]
                    total_count = (
                        counter.silence_following_counts[(w, p)]
                        + counter.non_silence_following_counts[(w, p)]
                    )
                    w_p_silence_count = count + (silence_probability * lambda_2)
                    prob = format_probability(w_p_silence_count / (total_count + lambda_2))
                    silence_probabilities[(w, p)] = prob
                    if w not in {initial_key[0], final_key[0], self.silence_word}:
                        pron_mapping[(w, p)]["silence_after_probability"] = prob
                lambda_3 = 2
                bar_count_silence_wp = collections.defaultdict(float)
                bar_count_non_silence_wp = collections.defaultdict(float)
                for (w_p1, w_p2), counts in counter.ngram_counts.items():
                    if w_p1 not in silence_probabilities:
                        silence_prob = 0.01
                    else:
                        silence_prob = silence_probabilities[w_p1]
                    total_count = counts["silence"] + counts["non_silence"]
                    bar_count_silence_wp[w_p2] += total_count * silence_prob
                    bar_count_non_silence_wp[w_p2] += total_count * (1 - silence_prob)
                for w, p, _ in pronunciations:
                    if w in {initial_key[0], final_key[0], self.silence_word}:
                        continue
                    silence_count = counter.silence_before_counts[(w, p)]
                    non_silence_count = counter.non_silence_before_counts[(w, p)]
                    pron_mapping[(w, p)]["silence_before_correction"] = format_correction(
                        (silence_count + lambda_3) / (bar_count_silence_wp[(w, p)] + lambda_3)
                    )

                    pron_mapping[(w, p)]["non_silence_before_correction"] = format_correction(
                        (non_silence_count + lambda_3)
                        / (bar_count_non_silence_wp[(w, p)] + lambda_3)
                    )
                session.bulk_update_mappings(Pronunciation, pron_mapping.values())
                session.flush()
                initial_silence_count = counter.silence_before_counts[initial_key] + (
                    silence_probability * lambda_2
                )
                initial_non_silence_count = counter.non_silence_before_counts[initial_key] + (
                    (1 - silence_probability) * lambda_2
                )
                initial_silence_probability = format_probability(
                    initial_silence_count / (initial_silence_count + initial_non_silence_count)
                )

                final_silence_correction = format_correction(
                    (counter.silence_before_counts[final_key] + lambda_3)
                    / (bar_count_silence_wp[final_key] + lambda_3)
                )

                final_non_silence_correction = format_correction(
                    (counter.non_silence_before_counts[final_key] + lambda_3)
                    / (bar_count_non_silence_wp[final_key] + lambda_3)
                )
                initial_silence_prob_sum += initial_silence_probability
                final_silence_correction_sum += final_silence_correction
                final_non_silence_correction_sum += final_non_silence_correction
                dictionary_mappings.append(
                    {
                        "id": d_id,
                        "silence_probability": silence_probability,
                        "initial_silence_probability": initial_silence_probability,
                        "final_silence_correction": final_silence_correction,
                        "final_non_silence_correction": final_non_silence_correction,
                    }
                )
            if compute_silence_probabilities:
                self.silence_probability = silence_prob_sum / self.num_dictionaries
                self.initial_silence_probability = initial_silence_prob_sum / self.num_dictionaries
                self.final_silence_correction = (
                    final_silence_correction_sum / self.num_dictionaries
                )
                self.final_non_silence_correction = (
                    final_non_silence_correction_sum / self.num_dictionaries
                )
            session.bulk_update_mappings(Dictionary, dictionary_mappings)
            session.commit()
        self.log_debug(f"Calculating pronunciation probabilities took {time.time() - begin}")

    def _collect_alignments(self):
        """
        Process alignment archives to extract word or phone alignments

        See Also
        --------
        :class:`~montreal_forced_aligner.alignment.multiprocessing.AlignmentExtractionFunction`
            Multiprocessing function for extracting alignments
        :meth:`.CorpusAligner.word_alignment_arguments`
            Arguments for word CTMs
        :meth:`.CorpusAligner.phone_alignment_arguments`
            Arguments for phone alignment
        """
        self.log_info("Collecting phone and word alignments from alignment lattices...")
        jobs = self.alignment_extraction_arguments()  # Phone CTM jobs
        with self.session() as session, tqdm.tqdm(
            total=self.num_current_utterances, disable=getattr(self, "quiet", False)
        ) as pbar:
            phone_interval_mappings = []
            word_interval_mappings = []
            if self.use_mp:
                error_dict = {}
                return_queue = mp.Queue()
                stopped = Stopped()
                procs = []
                for i, args in enumerate(jobs):
                    function = AlignmentExtractionFunction(args)
                    p = KaldiProcessWorker(i, return_queue, function, stopped)
                    procs.append(p)
                    p.start()
                while True:
                    try:
                        result = return_queue.get(timeout=1)
                        if isinstance(result, Exception):
                            error_dict[getattr(result, "job_name", 0)] = result
                            continue
                        if stopped.stop_check():
                            continue
                    except Empty:
                        for proc in procs:
                            if not proc.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    utterance, word_intervals, phone_intervals = result
                    for interval in phone_intervals:
                        phone_interval_mappings.append(
                            {
                                "begin": interval.begin,
                                "end": interval.end,
                                "label": interval.label,
                                "utterance_id": utterance,
                            }
                        )
                    for interval in word_intervals:
                        word_interval_mappings.append(
                            {
                                "begin": interval.begin,
                                "end": interval.end,
                                "label": interval.label,
                                "utterance_id": utterance,
                            }
                        )

                    pbar.update(1)
                for p in procs:
                    p.join()
                if error_dict:
                    for v in error_dict.values():
                        raise v
            else:
                for args in jobs:
                    function = AlignmentExtractionFunction(args)
                    for utterance, word_intervals, phone_intervals in function.run():

                        for interval in phone_intervals:
                            phone_interval_mappings.append(
                                {
                                    "begin": interval.begin,
                                    "end": interval.end,
                                    "label": interval.label,
                                    "utterance_id": utterance,
                                }
                            )
                        for interval in word_intervals:
                            word_interval_mappings.append(
                                {
                                    "begin": interval.begin,
                                    "end": interval.end,
                                    "label": interval.label,
                                    "utterance_id": utterance,
                                }
                            )

                        pbar.update(1)
            self.alignment_done = True
            session.query(Corpus).update({"alignment_done": True})
            session.bulk_insert_mappings(
                PhoneInterval, phone_interval_mappings, return_defaults=False, render_nulls=True
            )
            session.bulk_insert_mappings(
                WordInterval, word_interval_mappings, return_defaults=False, render_nulls=True
            )
            session.commit()

    def collect_alignments(self) -> None:
        """
        Collect word and phone alignments from alignment archives
        """
        if self.alignment_done:
            if self.export_output_directory is not None:
                self.export_textgrids()
            return
        self._collect_alignments()

    def export_textgrids(
        self, output_format: str = TextFileType.TEXTGRID.value, include_original_text: bool = False
    ) -> None:
        """
        Exports alignments to TextGrid files

        See Also
        --------
        :class:`~montreal_forced_aligner.alignment.multiprocessing.ExportTextGridProcessWorker`
            Multiprocessing helper function for TextGrid export
        :meth:`.CorpusAligner.export_textgrid_arguments`
            Job method for TextGrid export

        Parameters
        ----------
        output_format: str, optional
            Format to save alignments, one of 'long_textgrids' (the default), 'short_textgrids', or 'json', passed to praatio
        """
        if not self.alignment_done:
            self._collect_alignments()
        begin = time.time()
        error_dict = {}

        with tqdm.tqdm(total=self.num_files, disable=getattr(self, "quiet", False)) as pbar:

            with self.session() as session:
                files = (
                    session.query(
                        File.id,
                        File.name,
                        File.relative_path,
                        SoundFile.duration,
                        TextFile.text_file_path,
                    )
                    .join(File.sound_file)
                    .join(File.text_file)
                )
                if self.use_mp and self.num_jobs > 1:
                    stopped = Stopped()

                    finished_adding = Stopped()
                    for_write_queue = mp.Queue()
                    return_queue = mp.Queue()
                    export_args = self.export_textgrid_arguments(
                        output_format, include_original_text
                    )
                    exported_file_count = Counter()
                    export_procs = []
                    self.db_engine.dispose()
                    for j in self.jobs:
                        if not j.has_data:
                            continue
                        export_proc = ExportTextGridProcessWorker(
                            self.db_path,
                            for_write_queue,
                            return_queue,
                            stopped,
                            finished_adding,
                            export_args[j.name],
                            exported_file_count,
                        )
                        export_proc.start()
                        export_procs.append(export_proc)
                    try:
                        for args in files:
                            for_write_queue.put(args)
                        time.sleep(1)
                        finished_adding.stop()
                        while True:
                            try:
                                result = return_queue.get(timeout=1)
                                if isinstance(result, AlignmentExportError):
                                    error_dict[getattr(result, "path", 0)] = result
                                    continue
                                if self.stopped.stop_check():
                                    continue
                            except Empty:
                                for proc in export_procs:
                                    if not proc.finished_processing.stop_check():
                                        break
                                else:
                                    break
                                continue
                            if isinstance(result, int):
                                pbar.update(1)
                    except Exception:
                        stopped.stop()
                        raise
                    finally:

                        for i in range(self.num_jobs):
                            export_procs[i].join()
                else:
                    self.log_debug("Not using multiprocessing for TextGrid export")

                    for file_id, name, relative_path, duration, text_file_path in files:
                        output_path = construct_output_path(
                            name,
                            relative_path,
                            self.export_output_directory,
                            text_file_path,
                            output_format,
                        )
                        utterances = (
                            session.query(Utterance)
                            .options(
                                joinedload(Utterance.speaker, innerjoin=True).load_only(
                                    Speaker.name
                                ),
                                selectinload(Utterance.phone_intervals),
                                selectinload(Utterance.word_intervals),
                            )
                            .filter(Utterance.file_id == file_id)
                        )
                        data = {}
                        for utt in utterances:
                            if utt.speaker.name not in data:
                                data[utt.speaker.name] = {"words": [], "phones": []}
                            for wi in utt.word_intervals:
                                data[utt.speaker.name]["words"].append(
                                    CtmInterval(wi.begin, wi.end, wi.label, utt.id)
                                )

                            for pi in utt.phone_intervals:
                                data[utt.speaker.name]["phones"].append(
                                    CtmInterval(pi.begin, pi.end, pi.label, utt.id)
                                )
                        export_textgrid(
                            data,
                            output_path,
                            duration,
                            self.frame_shift,
                            output_format=output_format,
                        )
                        pbar.update(1)

        if error_dict:
            self.log_warning(
                f"There were {len(error_dict)} errors encountered in generating TextGrids. "
                f"Check {os.path.join(self.export_output_directory, 'output_errors.txt')} "
                f"for more details"
            )
        output_textgrid_writing_errors(self.export_output_directory, error_dict)
        self.log_info(f"Finished exporting TextGrids to {self.export_output_directory}!")
        self.log_debug(f"Exported TextGrids in a total of {time.time() - begin} seconds")

    def export_files(
        self,
        output_directory: str,
        output_format: Optional[str] = None,
        include_original_text: bool = False,
    ) -> None:
        """
        Export a TextGrid file for every sound file in the dataset

        Parameters
        ----------
        output_directory: str
            Directory to save to
        output_format: str, optional
            Format to save alignments, one of 'long_textgrids' (the default), 'short_textgrids', or 'json', passed to praatio
        """
        if output_format is None:
            output_format = TextFileType.TEXTGRID.value
        self.export_output_directory = output_directory

        self.log_info(f"Exporting TextGrids to {self.export_output_directory}...")
        os.makedirs(self.export_output_directory, exist_ok=True)
        self.export_textgrids(output_format, include_original_text)

    def evaluate_alignments(
        self,
        mapping: Optional[Dict[str, str]] = None,
        output_directory: Optional[str] = None,
    ) -> None:
        """
        Evaluate alignments against a reference directory

        Parameters
        ----------
        mapping: dict[str, Union[str, list[str]]], optional
            Mapping between phones that should be considered equal across different phone set types
        output_directory: str, optional
            Directory to save results, if not specified, it will be saved in the log directory
        """
        begin = time.time()
        if output_directory:
            csv_path = os.path.join(output_directory, "alignment_evaluation.csv")
        else:
            csv_path = os.path.join(self.working_log_directory, "alignment_evaluation.csv")
        csv_header = [
            "file",
            "begin",
            "end",
            "speaker",
            "duration",
            "normalized_text",
            "oovs",
            "reference_phone_count",
            "alignment_score",
            "phone_error_rate",
            "alignment_log_likelihood",
            "word_count",
            "oov_count",
        ]

        score_count = 0
        score_sum = 0
        phone_edit_sum = 0
        phone_length_sum = 0
        if self.alignment_evaluation_done:
            self.log_info("Exporting saved evaluation...")
            with self.session() as session, open(csv_path, "w", encoding="utf8", newline="") as f:
                writer = csv.DictWriter(f, fieldnames=csv_header)
                writer.writeheader()
                bn = DictBundle(
                    "evaluation_data",
                    File.c.name.label("file"),
                    Utterance.begin,
                    Utterance.end,
                    Speaker.c.name.label("speaker"),
                    Utterance.duration,
                    Utterance.normalized_text,
                    Utterance.oovs,
                    sqlalchemy.func.count(Utterance.reference_phone_intervals).label(
                        "reference_phone_count"
                    ),
                    Utterance.alignment_score,
                    Utterance.phone_error_rate,
                    Utterance.alignment_log_likelihood,
                )
                utterances = (
                    session.query(bn)
                    .join(Utterance.speaker)
                    .join(Utterance.file)
                    .group_by(Utterance.id)
                    .join(Utterance.reference_phone_intervals)
                )
                for line in utterances:
                    data = line["evaluation_data"]
                    data["word_count"] = len(data["normalized_text"].split())
                    data["oov_count"] = len(data["oovs"].split())
                    phone_error_rate = data["phone_error_rate"]
                    reference_phone_count = data["reference_phone_count"]
                    if data["alignment_score"] is not None:
                        score_count += 1
                        score_sum += data["alignment_score"]
                    phone_edit_sum += int(phone_error_rate * reference_phone_count)
                    phone_length_sum += reference_phone_count
                    writer.writerow(data)
        else:
            # Set up
            self.log_info("Evaluating alignments...")
            self.log_debug(f"Mapping: {mapping}")
            update_mappings = []
            indices = []
            to_comp = []
            score_func = functools.partial(
                align_phones,
                silence_phone=self.optional_silence_phone,
                ignored_phones={self.oov_phone},
                custom_mapping=mapping,
            )
            with self.session() as session:
                unaligned_utts = []
                utterances = session.query(Utterance).options(
                    subqueryload(Utterance.reference_phone_intervals),
                    subqueryload(Utterance.phone_intervals),
                    joinedload(Utterance.file, innerjoin=True),
                    joinedload(Utterance.speaker, innerjoin=True),
                )
                for u in utterances:
                    reference_phone_count = len(u.reference_phone_intervals)
                    if not reference_phone_count:
                        continue
                    if u.alignment_log_likelihood is None:  # couldn't be aligned
                        phone_error_rate = reference_phone_count
                        unaligned_utts.append(u)
                        update_mappings.append(
                            {
                                "id": u.id,
                                "alignment_score": None,
                                "phone_error_rate": phone_error_rate,
                            }
                        )
                        continue
                    reference_phone_labels = [x.as_ctm() for x in u.reference_phone_intervals]
                    phone_labels = [x.as_ctm() for x in u.phone_intervals]
                    indices.append(u)
                    to_comp.append((reference_phone_labels, phone_labels))

                with mp.Pool(self.num_jobs) as pool, open(
                    csv_path, "w", encoding="utf8", newline=""
                ) as f:
                    writer = csv.DictWriter(f, fieldnames=csv_header)
                    writer.writeheader()
                    gen = pool.starmap(score_func, to_comp)
                    for u in unaligned_utts:
                        word_count = len(u.normalized_text.split())
                        oov_count = len(u.oovs.split())
                        reference_phone_count = len(u.reference_phone_intervals)
                        writer.writerow(
                            {
                                "file": u.file_name,
                                "begin": u.begin,
                                "end": u.end,
                                "speaker": u.speaker_name,
                                "duration": u.duration,
                                "normalized_text": u.normalized_text,
                                "oovs": u.oovs,
                                "reference_phone_count": reference_phone_count,
                                "alignment_score": None,
                                "phone_error_rate": reference_phone_count,
                                "alignment_log_likelihood": None,
                                "word_count": word_count,
                                "oov_count": oov_count,
                            }
                        )
                    for i, (score, phone_error_rate) in enumerate(gen):
                        if score is None:
                            continue
                        u = indices[i]

                        word_count = len(u.normalized_text.split())
                        oov_count = len(u.oovs.split())
                        reference_phone_count = len(u.reference_phone_intervals)
                        update_mappings.append(
                            {
                                "id": u.id,
                                "alignment_score": score,
                                "phone_error_rate": phone_error_rate,
                            }
                        )
                        writer.writerow(
                            {
                                "file": u.file_name,
                                "begin": u.begin,
                                "end": u.end,
                                "speaker": u.speaker_name,
                                "duration": u.duration,
                                "normalized_text": u.normalized_text,
                                "oovs": u.oovs,
                                "reference_phone_count": reference_phone_count,
                                "alignment_score": score,
                                "phone_error_rate": phone_error_rate,
                                "alignment_log_likelihood": u.alignment_log_likelihood,
                                "word_count": word_count,
                                "oov_count": oov_count,
                            }
                        )
                        score_count += 1
                        score_sum += score
                        phone_edit_sum += int(phone_error_rate * reference_phone_count)
                        phone_length_sum += reference_phone_count
                session.bulk_update_mappings(Utterance, update_mappings)
                session.query(Corpus).update({"alignment_evaluation_done": True})
                session.commit()
        self.log_info(f"Average overlap score: {score_sum/score_count}")
        self.log_info(f"Average phone error rate: {phone_edit_sum/phone_length_sum}")
        self.log_debug(f"Alignment evaluation took {time.time()-begin} seconds")



from __future__ import annotations

import multiprocessing as mp
import os
import queue
import random
import subprocess
import time
from typing import Dict

import dataclassy
import numpy
import tqdm

from montreal_forced_aligner.abc import MetaDict, TopLevelMfaWorker
from montreal_forced_aligner.data import WordType
from montreal_forced_aligner.db import Pronunciation, Word
from montreal_forced_aligner.dictionary.multispeaker import MultispeakerDictionaryMixin
from montreal_forced_aligner.g2p.generator import PyniniValidator
from montreal_forced_aligner.g2p.trainer import G2PTrainer
from montreal_forced_aligner.models import G2PModel
from montreal_forced_aligner.utils import Stopped, thirdparty_binary

try:
    import pynini
    import pywrapfst
    from pynini.lib import rewrite

    G2P_DISABLED = False

except ImportError:
    pynini = None
    pywrapfst = None
    rewrite = None

    G2P_DISABLED = True


__all__ = ["PhonetisaurusTrainerMixin", "PhonetisaurusTrainer"]


@dataclassy.dataclass(slots=True)
class LabelData:
    """Label data class for penalizing alignments based on the size of their left/right side"""

    max: int
    tot: int
    lhs: int
    rhs: int
    lhsE: bool
    rhsE: bool


@dataclassy.dataclass(slots=True)
class MaximizationArguments:
    """Arguments for the maximization worker"""

    far_path: str
    alignment_model: Dict[int, pynini.Weight]
    penalize_em: bool
    penalties: Dict[int, LabelData]


@dataclassy.dataclass(slots=True)
class AlignmentInitArguments:
    """Arguments for the alignment initialization worker"""

    far_path: str
    deletions: bool
    insertions: bool
    restrict: bool
    phone_order: int
    grapheme_order: int
    s1s2_sep: str
    seq_sep: str
    skip: str


class AlignmentInitWorker(mp.Process):
    """
    Multiprocessing worker that initializes alignment FSTs for a subset of the data

    Parameters
    ----------
    job_q: mp.Queue
        Queue of grapheme-phoneme transcriptions to process
    return_queue: mp.Queue
        Queue to return data
    stopped: Stopped
        Stop check
    finished_adding: Stopped
        Check for whether the job queue is done
    symbol_dict: dict
        Symbol to integer ID mapping dictionary
    next_symbol: mp.Value
        Integer value to use for the next symbol
    lock: mp.Lock
        Lock to use for data shared across processes
    args: :class:`~montreal_forced_aligner.g2p.phonetisaurus_trainer.AlignmentInitArguments`
        Arguments for initialization
    """

    def __init__(
        self,
        job_q: mp.Queue,
        return_queue: mp.Queue,
        stopped: Stopped,
        finished_adding: Stopped,
        symbol_dict: dict,
        next_symbol: mp.Value,
        lock: mp.Lock,
        args: AlignmentInitArguments,
    ):
        mp.Process.__init__(self)
        self.job_q = job_q
        self.return_queue = return_queue
        self.stopped = stopped
        self.symbol_cache = {}
        self.symbol_dict = symbol_dict
        self.next_symbol: mp.Value = next_symbol
        self.lock = lock
        self.finished = Stopped()
        self.finished_adding = finished_adding
        self.deletions = args.deletions
        self.insertions = args.insertions
        self.restrict = args.restrict
        self.phone_order = args.phone_order
        self.grapheme_order = args.grapheme_order
        self.s1s2_sep = args.s1s2_sep
        self.seq_sep = args.seq_sep
        self.skip = args.skip
        self.far_path = args.far_path

    def look_up_symbol(self, symbol: str) -> int:
        """
        Look up a symbol based on the process's symbol cache and symbol table

        Parameters
        ----------
        symbol: str
            Symbol to look up

        Returns
        -------
        int
            Symbol ID in table
        """
        if symbol not in self.symbol_cache:
            with self.lock:
                if symbol not in self.symbol_dict:
                    self.symbol_dict[symbol] = self.next_symbol.value
                    self.next_symbol.value += 1
                self.symbol_cache[symbol] = self.symbol_dict[symbol]
        return self.symbol_cache[symbol]

    def run(self) -> None:
        """Run the function"""
        current_index = 0
        far_writer = pywrapfst.FarWriter.create(self.far_path, arc_type="log")
        while True:
            try:
                graphemes, phones = self.job_q.get(timeout=1)
            except queue.Empty:
                if self.finished_adding.stop_check():
                    break
                continue
            if self.stopped.stop_check():
                continue
            try:
                key = f"{current_index:08x}"
                fst = pynini.Fst(arc_type="log")
                initial_weight = pywrapfst.Weight(fst.weight_type(), 99)
                final_state = ((len(graphemes) + 1) * (len(phones) + 1)) - 1
                for _ in range(final_state + 1):
                    fst.add_state()
                for i in range(len(graphemes) + 1):
                    for j in range(len(phones) + 1):
                        istate = i * (len(phones) + 1) + j
                        if self.deletions:
                            for phone_range in range(1, self.phone_order + 1):
                                if j + phone_range <= len(phones):
                                    subseq_phones = phones[j : j + phone_range]
                                    symbol = self.look_up_symbol(
                                        self.s1s2_sep.join(
                                            [self.skip, self.seq_sep.join(subseq_phones)]
                                        )
                                    )
                                    ostate = i * (len(phones) + 1) + (j + phone_range)
                                    fst.add_arc(
                                        istate,
                                        pywrapfst.Arc(symbol, symbol, initial_weight, ostate),
                                    )
                        if self.insertions:
                            for k in range(1, self.grapheme_order + 1):
                                if i + k <= len(graphemes):
                                    subseq_graphemes = graphemes[i : i + k]
                                    symbol = self.look_up_symbol(
                                        self.s1s2_sep.join(
                                            [self.seq_sep.join(subseq_graphemes), self.skip]
                                        )
                                    )
                                    ostate = (i + k) * (len(phones) + 1) + j
                                    fst.add_arc(
                                        istate,
                                        pywrapfst.Arc(symbol, symbol, initial_weight, ostate),
                                    )

                        for grapheme_range in range(1, self.grapheme_order + 1):
                            for phone_range in range(1, self.phone_order + 1):
                                if i + grapheme_range <= len(graphemes) and j + phone_range <= len(
                                    phones
                                ):
                                    if self.restrict and grapheme_range > 1 and phone_range > 1:
                                        continue
                                    subseq_phones = phones[j : j + phone_range]
                                    phone_string = self.seq_sep.join(subseq_phones)
                                    subseq_graphemes = graphemes[i : i + grapheme_range]
                                    grapheme_string = self.seq_sep.join(subseq_graphemes)
                                    symbol = self.look_up_symbol(
                                        self.s1s2_sep.join([grapheme_string, phone_string])
                                    )
                                    ostate = (i + grapheme_range) * (len(phones) + 1) + (
                                        j + phone_range
                                    )
                                    weight = pywrapfst.Weight(
                                        fst.weight_type(), float(grapheme_range * phone_range)
                                    )
                                    fst.add_arc(
                                        istate, pywrapfst.Arc(symbol, symbol, weight, ostate)
                                    )
                fst.set_start(0)
                fst.set_final(final_state, pywrapfst.Weight.one(fst.weight_type()))

                if not self.insertions or not self.deletions:
                    fst = pynini.connect(fst)
                far_writer[key] = fst
                self.return_queue.put(fst)
                current_index += 1
            except Exception as e:  # noqa
                self.stopped.stop()
                self.return_queue.put(e)
                raise
        self.finished.stop()
        return


class MaximizationWorker(mp.Process):
    """
    Multiprocessing worker that runs the maximization step of training for a subset of the data

    Parameters
    ----------
    return_queue: mp.Queue
        Queue to return data
    stopped: Stopped
        Stop check
    args: :class:`~montreal_forced_aligner.g2p.phonetisaurus_trainer.MaximizationArguments`
        Arguments for maximization
    """

    def __init__(self, return_queue: mp.Queue, stopped: Stopped, args: MaximizationArguments):
        mp.Process.__init__(self)
        self.return_queue = return_queue
        self.stopped = stopped
        self.finished = Stopped()
        self.penalize_em = args.penalize_em
        self.alignment_model = args.alignment_model
        self.penalties = args.penalties
        self.far_path = args.far_path

    def run(self) -> None:
        """Run the function"""
        zero = pynini.Weight.zero("log")
        far_reader = pywrapfst.FarReader.open(self.far_path)
        far_writer = pywrapfst.FarWriter.create(self.far_path + ".temp", arc_type="log")
        while not far_reader.done():
            if self.stopped.stop_check():
                break
            key = far_reader.get_key()
            fst = far_reader.get_fst()
            for state_id in fst.states():
                maiter = fst.mutable_arcs(state_id)
                while not maiter.done():
                    arc = maiter.value()
                    if not self.penalize_em:
                        arc.weight = self.alignment_model[arc.ilabel]
                    else:
                        label_data = self.penalties[arc.ilabel]
                        if label_data.lhs > 1 and label_data.rhs > 1:
                            arc.weight = pynini.Weight(fst.weight_type(), 99)
                        elif not label_data.lhsE and not label_data.rhsE:
                            arc.weight = pynini.Weight(
                                fst.weight_type(), float(arc.weight) * label_data.tot
                            )
                    if arc.weight == zero:
                        arc.weight = pynini.Weight(fst.weight_type(), 99)
                    arc = pywrapfst.Arc(arc.ilabel, arc.olabel, arc.weight, arc.nextstate)
                    maiter.set_value(arc)
                    next(maiter)
            far_writer[key] = fst
            next(far_reader)
            self.return_queue.put(1)
        os.remove(self.far_path)
        os.rename(self.far_path + ".temp", self.far_path)
        self.finished.stop()


class ExpectationWorker(mp.Process):
    """
    Multiprocessing worker that runs the expectation step of training for a subset of the data

    Parameters
    ----------
    far_path: str
        Path to FST archive file
    return_queue: mp.Queue
        Queue to return data
    stopped: Stopped
        Stop check
    """

    def __init__(
        self,
        far_path: str,
        return_queue: mp.Queue,
        stopped: Stopped,
    ):
        mp.Process.__init__(self)
        self.far_path = far_path
        self.return_queue = return_queue
        self.stopped = stopped
        self.finished = Stopped()

    def run(self) -> None:
        """Run the function"""
        far_reader = pywrapfst.FarReader.open(self.far_path)
        while not far_reader.done():
            if self.stopped.stop_check():
                break
            key = far_reader.get_key()
            fst = far_reader.get_fst()
            data = {}
            try:
                fst = pynini.Fst.read_from_string(fst.write_to_string())
                alpha = pynini.shortestdistance(fst)
                beta = pynini.shortestdistance(fst, reverse=True)
                for state_id in fst.states():
                    for arc in fst.arcs(state_id):
                        gamma = pynini.divide(
                            pynini.times(
                                pynini.times(alpha[state_id], arc.weight), beta[arc.nextstate]
                            ),
                            beta[0],
                        )
                        if float(gamma) != numpy.inf:
                            if arc.ilabel not in data:
                                data[arc.ilabel] = 0
                            data[arc.ilabel] += float(gamma)
                self.return_queue.put((key, data))
                next(far_reader)
            except Exception as e:  # noqa
                self.stopped.stop()
                self.return_queue.put(e)
                raise
        self.finished.stop()
        return


class PhonetisaurusTrainerMixin:
    """
    Mixin class for training Phonetisaurus style models

    Parameters
    ----------
    order: int
        Order of the ngram model, defaults to 7
    random_starts: int
        Number of random starts to use in initialization, defaults to 25
    seed: int
        Seed for randomization, defaults to 1917
    delta: float
        Comparison/quantization delta for Baum-Welch training, defaults to 1/1024
    alpha: float
        Step size reduction power parameter for Baum-Welch training;
        full standard batch EM is run (not stepwise) if set to 0, defaults to 1.0
    batch_size:int
        Batch size for Baum-Welch training, defaults to 200
    num_iterations:int
        Maximum number of iterations to use in Baum-Welch training, defaults to 10
    smoothing_method:str
        Smoothing method for the ngram model, defaults to "kneser_ney"
    pruning_method:str
        Pruning method for pruning the ngram model, defaults to "relative_entropy"
    model_size: int
        Target number of ngrams for pruning, defaults to 1000000
    insertions: bool
        Flag for whether to allow for insertions, default True
    deletions: bool
        Flag for whether to allow for deletions, default True
    grapheme_order: int
        Maximum number of graphemes to map to single phones
    phone_order: int
        Maximum number of phones to map to single graphemes
    fst_default_cache_gc: str
        String to pass to OpenFst binaries for GC behavior
    fst_default_cache_gc_limit: str
        String to pass to OpenFst binaries for GC behavior
    """

    def __init__(
        self,
        order: int = 8,
        num_iterations: int = 10,
        smoothing_method: str = "kneser_ney",
        pruning_method: str = "relative_entropy",
        model_size: int = 1000000,
        insertions: bool = True,
        deletions: bool = True,
        grapheme_order: int = 2,
        phone_order: int = 2,
        fst_default_cache_gc="",
        fst_default_cache_gc_limit="",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not hasattr(self, "_data_source"):
            self._data_source = None
        self.order = order
        self.num_iterations = num_iterations
        self.smoothing_method = smoothing_method
        self.pruning_method = pruning_method
        self.model_size = model_size
        self.insertions = insertions
        self.deletions = deletions
        self.fst_default_cache_gc = fst_default_cache_gc
        self.fst_default_cache_gc_limit = fst_default_cache_gc_limit
        self.grapheme_order = grapheme_order
        self.phone_order = phone_order
        self.seq_sep = "|"
        self.s1s2_sep = "}"
        self.skip = "_"
        self.eps = "<eps>"
        self.restrict = True
        self.penalize_em = True
        self.penalize = True
        self.g2p_num_training_pronunciations = 0

    @property
    def architecture(self) -> str:
        """Phonetisaurus"""
        return "phonetisaurus"

    def initialize_alignments(self) -> None:
        """
        Initialize alignment FSTs for training

        """
        self.symbol_table = pynini.SymbolTable()
        self.symbol_table.add_symbol(self.eps)
        self.symbol_table.add_symbol(self.skip)
        self.symbol_table.add_symbol(f"{self.seq_sep}_{self.seq_sep}")
        self.symbol_table.add_symbol(self.s1s2_sep)
        model_params = [
            "true" if self.deletions else "false",
            "true" if self.insertions else "false",
            str(self.grapheme_order),
            str(self.phone_order),
        ]
        self.symbol_table.add_symbol("_".join(model_params))
        self.alignment_model: Dict[int, pynini.Weight] = {}
        self.prev_alignment_model: Dict[int, pynini.Weight] = {}
        self.penalties: Dict[int, LabelData] = {}
        self.total = pynini.Weight.zero("log")
        self.prev_total = pynini.Weight.zero("log")

        self.fsas = []
        self.log_info("Creating alignment FSTs...")
        with mp.Manager() as manager:
            mp_symbol_dict = manager.dict()
            lock = mp.Lock()
            next_symbol = mp.Value("i", self.symbol_table.num_symbols())
            for i in range(self.symbol_table.num_symbols()):
                sym = self.symbol_table.find(i)
                mp_symbol_dict[sym] = i
            job_queue = mp.Queue()
            return_queue = mp.Queue()
            stopped = Stopped()
            finished_adding = Stopped()
            procs = []
            for i in range(self.num_jobs):
                args = AlignmentInitArguments(
                    os.path.join(self.working_directory, f"{i}.far"),
                    self.deletions,
                    self.insertions,
                    self.restrict,
                    self.phone_order,
                    self.grapheme_order,
                    self.s1s2_sep,
                    self.seq_sep,
                    self.skip,
                )
                procs.append(
                    AlignmentInitWorker(
                        job_queue,
                        return_queue,
                        stopped,
                        finished_adding,
                        mp_symbol_dict,
                        next_symbol,
                        lock,
                        args,
                    )
                )
                procs[i].start()
            self.g2p_num_training_pronunciations = 0
            for word, pronunciations in self.g2p_training_dictionary.items():
                graphemes = list(word)
                for p in pronunciations:
                    phones = p.split()
                    job_queue.put((graphemes, phones))
                    self.g2p_num_training_pronunciations += 1
            finished_adding.stop()
            error_list = []
            with tqdm.tqdm(
                total=self.g2p_num_training_pronunciations, disable=getattr(self, "quiet", False)
            ) as pbar:
                while True:
                    try:
                        fst = return_queue.get(timeout=1)
                        if isinstance(fst, Exception):
                            error_list.append(fst)
                            continue
                        if stopped.stop_check():
                            continue
                    except queue.Empty:
                        for p in procs:
                            if not p.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    for state_id in fst.states():
                        for arc in fst.arcs(state_id):
                            if arc.ilabel not in self.prev_alignment_model:
                                self.prev_alignment_model[arc.ilabel] = arc.weight
                                sym = mp_symbol_dict[arc.ilabel]
                                d = sym.find("}")
                                c = sym.find("|")
                                left_side_count = 1
                                right_side_count = 1
                                if c != -1:
                                    if c < d:
                                        left_side_count += 1
                                    else:
                                        right_side_count += 1
                                max_count = max(left_side_count, right_side_count)
                                self.penalties[arc.ilabel] = LabelData(
                                    tot=left_side_count + right_side_count,
                                    max=max_count,
                                    lhs=left_side_count,
                                    rhs=right_side_count,
                                    lhsE=False,
                                    rhsE=False,
                                )
                            else:
                                self.prev_alignment_model[arc.ilabel] = pynini.plus(
                                    self.prev_alignment_model[arc.ilabel], arc.weight
                                )
                            self.total = pynini.plus(self.total, arc.weight)
                    pbar.update(1)
            for p in procs:
                p.join()
            for k in sorted(mp_symbol_dict.keys(), key=lambda x: mp_symbol_dict[x]):
                if self.symbol_table.find(k) == pynini.NO_SYMBOL:
                    self.symbol_table.add_symbol(k)
        if error_list:
            for v in error_list:
                raise v

    def maximization(self) -> float:
        """
        Run the maximization step for training

        Returns
        -------
        float
            Current iteration's score
        """
        self.log_info("Performing maximization step...")
        cond = False
        change = abs(float(self.total) - float(self.prev_total))
        zero = pynini.Weight.zero("log")
        if not cond:
            self.prev_total = self.total
            for ilabel, weight in self.prev_alignment_model.items():
                self.alignment_model[ilabel] = pynini.divide(weight, self.total)
                self.prev_alignment_model[ilabel] = zero
        return_queue = mp.Queue()
        stopped = Stopped()
        procs = []
        for i in range(self.num_jobs):
            args = MaximizationArguments(
                os.path.join(self.working_directory, f"{i}.far"),
                self.alignment_model,
                self.penalize_em,
                self.penalties,
            )
            procs.append(MaximizationWorker(return_queue, stopped, args))
            procs[i].start()

        error_list = []
        with tqdm.tqdm(
            total=self.g2p_num_training_pronunciations, disable=getattr(self, "quiet", False)
        ) as pbar:
            while True:
                try:
                    result = return_queue.get(timeout=1)
                    if isinstance(result, Exception):
                        error_list.append(result)
                        continue
                    if stopped.stop_check():
                        continue
                except queue.Empty:
                    for p in procs:
                        if not p.finished.stop_check():
                            break
                    else:
                        break
                    continue
                pbar.update(1)
        for p in procs:
            p.join()

        if error_list:
            for v in error_list:
                raise v
        self.total = zero
        self.log_info(f"Maximization done! Change from last iteration was {change:.3f}")
        return change

    def expectation(self) -> None:
        """
        Run the expectation step for training
        """
        self.log_info("Performing expectation step...")
        return_queue = mp.Queue()
        stopped = Stopped()
        error_list = []
        procs = []
        for i in range(self.num_jobs):
            procs.append(
                ExpectationWorker(
                    os.path.join(self.working_directory, f"{i}.far"), return_queue, stopped
                )
            )
            procs[i].start()

        with tqdm.tqdm(
            total=self.g2p_num_training_pronunciations, disable=getattr(self, "quiet", False)
        ) as pbar:
            while True:
                try:
                    result = return_queue.get(timeout=1)
                    if isinstance(result, Exception):
                        error_list.append(result)
                        continue
                    if stopped.stop_check():
                        continue
                    index, data = result
                except queue.Empty:
                    for p in procs:
                        if not p.finished.stop_check():
                            break
                    else:
                        break
                    continue
                for ilabel, gamma in data.items():
                    gamma = pynini.Weight("log", gamma)
                    self.prev_alignment_model[ilabel] = pynini.plus(
                        self.prev_alignment_model[ilabel], gamma
                    )
                    self.total = pynini.plus(self.total, gamma)
                pbar.update(1)
        for p in procs:
            p.join()

        if error_list:
            for v in error_list:
                raise v
        self.log_info("Expectation done!")

    def train_ngram_model(self) -> None:
        """
        Train an ngram model on the aligned FSTs
        """
        if os.path.exists(self.fst_path):
            self.log_info("Ngram model already exists.")
            return
        self.log_info("Training ngram model...")
        with open(
            os.path.join(self.working_log_directory, "model.log"), "w", encoding="utf8"
        ) as logf:
            ngramcount_proc = subprocess.Popen(
                [
                    thirdparty_binary("ngramcount"),
                    "--require_symbols=false",
                    "--round_to_int",
                    f"--order={self.order}",
                    self.far_path,
                ],
                stderr=logf,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            ngramcount_proc.communicate()
            ngrammake_proc = subprocess.Popen(
                [
                    thirdparty_binary("ngrammake"),
                    f"--method={self.smoothing_method}",
                ],
                stdin=ngramcount_proc.stdout,
                stderr=logf,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            ngrammake_proc.communicate()

            ngramshrink_proc = subprocess.Popen(
                [
                    thirdparty_binary("ngramshrink"),
                    f"--method={self.pruning_method}",
                    f"--target_number_of_ngrams={self.model_size}",
                    "-",
                    self.ngram_path,
                ],
                stdin=ngrammake_proc.stdout,
                stderr=logf,
                env=os.environ,
            )
            ngramshrink_proc.communicate()

            ngram_fst = pynini.Fst.read(self.ngram_path)
            grapheme_symbols = pynini.SymbolTable()
            grapheme_symbols.add_symbol(self.eps)
            grapheme_symbols.add_symbol(self.seq_sep)
            grapheme_symbols.add_symbol(self.skip)
            phone_symbols = pynini.SymbolTable()
            phone_symbols.add_symbol(self.eps)
            phone_symbols.add_symbol(self.seq_sep)
            phone_symbols.add_symbol(self.skip)
            single_phone_symbols = pynini.SymbolTable()
            single_phone_symbols.add_symbol(self.eps)
            single_phone_fst = pynini.Fst()
            start_state = single_phone_fst.add_state()
            single_phone_fst.set_start(start_state)
            one = pynini.Weight.one(single_phone_fst.weight_type())
            single_phone_fst.set_final(start_state, one)
            current_ind = 1
            for state in ngram_fst.states():
                maiter = ngram_fst.mutable_arcs(state)
                while not maiter.done():
                    arc = maiter.value()
                    symbol = self.symbol_table.find(arc.ilabel)
                    try:
                        grapheme, phone = symbol.split(self.s1s2_sep)
                        g_symbol = grapheme_symbols.find(grapheme)
                        if g_symbol == pynini.NO_SYMBOL:
                            g_symbol = grapheme_symbols.add_symbol(grapheme)
                        p_symbol = phone_symbols.find(phone)
                        if p_symbol == pynini.NO_SYMBOL:
                            p_symbol = phone_symbols.add_symbol(phone)
                            singles = phone.split(self.seq_sep)
                            for i, s in enumerate(singles):
                                s_symbol = single_phone_symbols.find(s)

                                if s_symbol == pynini.NO_SYMBOL:
                                    s_symbol = single_phone_symbols.add_symbol(s)
                                if i == 0:
                                    single_start = start_state
                                else:
                                    single_start = current_ind
                                if i < len(singles) - 1:
                                    current_ind = single_phone_fst.add_state()
                                    end_state = current_ind
                                else:
                                    end_state = start_state
                                single_phone_fst.add_arc(
                                    single_start,
                                    pywrapfst.Arc(
                                        p_symbol if i == 0 else 0, s_symbol, one, end_state
                                    ),
                                )

                        arc = pywrapfst.Arc(g_symbol, p_symbol, arc.weight, arc.nextstate)
                        maiter.set_value(arc)
                    except ValueError:
                        if symbol == "<eps>":
                            arc = pywrapfst.Arc(0, 0, arc.weight, arc.nextstate)
                            maiter.set_value(arc)
                        else:
                            raise
                        pass
                    next(maiter)
            for i in range(grapheme_symbols.num_symbols()):
                sym = grapheme_symbols.find(i)
                if sym in {self.eps, self.seq_sep, self.skip}:
                    continue
                parts = sym.split(self.seq_sep)
                if len(parts) > 1:
                    for s in parts:
                        if grapheme_symbols.find(s) == pynini.NO_SYMBOL:
                            k = grapheme_symbols.add_symbol(s)
                            ngram_fst.add_arc(1, pywrapfst.Arc(k, 2, 99, 1))

            for i in range(phone_symbols.num_symbols()):
                sym = phone_symbols.find(i)
                if sym in {self.eps, self.seq_sep, self.skip}:
                    continue
                parts = sym.split(self.seq_sep)
                if len(parts) > 1:
                    for s in parts:
                        if phone_symbols.find(s) == pynini.NO_SYMBOL:
                            k = phone_symbols.add_symbol(s)
                            ngram_fst.add_arc(1, pywrapfst.Arc(2, k, 99, 1))
            single_phone_fst.set_input_symbols(phone_symbols)
            single_phone_fst.set_output_symbols(single_phone_symbols)
            ngram_fst.set_input_symbols(grapheme_symbols)
            ngram_fst.set_output_symbols(phone_symbols)
            single_ngram_fst = pynini.compose(ngram_fst, single_phone_fst)
            single_ngram_fst.set_input_symbols(grapheme_symbols)
            single_ngram_fst.set_output_symbols(single_phone_symbols)
            grapheme_symbols.write_text(self.grapheme_symbols_path)
            single_phone_symbols.write_text(self.phone_symbols_path)
            single_ngram_fst.write(self.fst_path)

    def train_alignments(self) -> None:
        """
        Run an Expectation-Maximization (EM) training on alignment FSTs to generate well-aligned FSTs for ngram modeling
        """
        if os.path.exists(self.far_path):
            self.log_info("Using existing alignments.")
            self.symbol_table = pynini.SymbolTable.read_text(self.alignment_symbols_path)
            return
        self.initialize_alignments()

        self.maximization()
        self.log_info("Training alignments...")
        for i in range(self.num_iterations):
            self.log_info(f"Iteration {i}")
            self.expectation()
            change = self.maximization()
            if change < 1e-10:
                break
        self.export_alignments()

    @property
    def data_directory(self) -> str:
        """Data directory for trainer"""
        return self.working_directory

    def train_iteration(self) -> None:
        """Train iteration, not used"""
        pass

    @property
    def workflow_identifier(self) -> str:
        """Identifier for Phonetisaurus G2P trainer"""
        return "phonetisaurus_train_g2p"

    @property
    def data_source_identifier(self) -> str:
        """Dictionary name"""
        return self._data_source

    def export_model(self, output_model_path: str) -> None:
        """
        Export G2P model to specified path

        Parameters
        ----------
        output_model_path:str
            Path to export model
        """
        directory, filename = os.path.split(output_model_path)
        basename, _ = os.path.splitext(filename)
        models_temp_dir = os.path.join(self.working_directory, "model_archive_temp")
        model = G2PModel.empty(basename, root_directory=models_temp_dir)
        model.add_meta_file(self)
        model.add_fst_model(self.working_directory)
        model.add_sym_path(self.working_directory)
        if directory:
            os.makedirs(directory, exist_ok=True)
        basename, _ = os.path.splitext(output_model_path)
        model.dump(basename)
        model.clean_up()
        # self.clean_up()
        self.log_info(f"Saved model to {output_model_path}")

    @property
    def alignment_model_path(self) -> str:
        """Path to store alignment model FST"""
        return os.path.join(self.working_directory, "align.fst")

    @property
    def ngram_path(self) -> str:
        """Path to store ngram model"""
        return os.path.join(self.working_directory, "ngram.fst")

    @property
    def fst_path(self) -> str:
        """Path to store final trained model"""
        return os.path.join(self.working_directory, "model.fst")

    @property
    def alignment_symbols_path(self) -> str:
        """Path to alignment symbol table"""
        return os.path.join(self.working_directory, "alignment.syms")

    @property
    def grapheme_symbols_path(self) -> str:
        """Path to final model's grapheme symbol table"""
        return os.path.join(self.working_directory, "graphemes.txt")

    @property
    def phone_symbols_path(self) -> str:
        """Path to final model's phone symbol table"""
        return os.path.join(self.working_directory, "phones.txt")

    @property
    def far_path(self) -> str:
        """Path to store final aligned FSTs"""
        return os.path.join(self.working_directory, "aligned.far")

    def export_alignments(self) -> None:
        """
        Combine alignment training archives to a final combined FST archive to train the ngram model
        """
        self.log_info("Exporting final alignments...")
        model = pynini.Fst(arc_type="log")
        model.add_state()
        model.set_start(0)
        model.set_final(0, pynini.Weight.one(model.arc_type()))
        for ilabel, weight in self.alignment_model.items():
            model.add_arc(0, pywrapfst.Arc(ilabel, ilabel, weight, 0))
        model.set_input_symbols(self.symbol_table)
        model.write(self.alignment_model_path)
        set_symbols = False
        self.symbol_table.write_text(self.alignment_symbols_path)
        far_writer = pywrapfst.FarWriter.create(self.far_path)
        zero = pynini.Weight.zero("log")
        one = pynini.Weight.one("log")
        index = 0
        with tqdm.tqdm(
            total=self.g2p_num_training_pronunciations, disable=getattr(self, "quiet", False)
        ) as pbar:
            for i in range(self.num_jobs):
                far_reader = pywrapfst.FarReader.open(
                    os.path.join(self.working_directory, f"{i}.far")
                )
                while not far_reader.done():
                    fst = far_reader.get_fst()
                    tfst = pynini.arcmap(
                        pynini.Fst.read_from_string(fst.write_to_string()), map_type="to_std"
                    )
                    if self.penalize:
                        for state in tfst.states():
                            maiter = tfst.mutable_arcs(state)
                            while not maiter.done():
                                arc = maiter.value()
                                ld = self.penalties[arc.ilabel]
                                if ld.lhs > 1 and ld.rhs > 1:
                                    arc.weight = pynini.Weight(tfst.weight_type(), 999)
                                else:
                                    arc.weight = pynini.Weight(
                                        tfst.weight_type(), float(arc.weight) * ld.max
                                    )
                                maiter.set_value(arc)
                                next(maiter)
                    tfst = rewrite.lattice_to_dfa(tfst, True, 4).project("output").rmepsilon()
                    lfst = pynini.arcmap(tfst, map_type="to_log")
                    pfst = pynini.push(lfst, reweight_type="to_final", push_weights=True)
                    for state in pfst.states():
                        if pfst.final(state) != zero:
                            pfst.set_final(state, one)
                    lattice = pynini.arcmap(pfst, map_type="to_std")

                    if not set_symbols:
                        lattice.set_input_symbols(self.symbol_table)
                        lattice.set_output_symbols(self.symbol_table)
                        set_symbols = True
                    key = f"{index:08x}"
                    far_writer[key] = lattice
                    pbar.update(1)
                    index += 1
                    next(far_reader)


class PhonetisaurusTrainer(
    MultispeakerDictionaryMixin, PhonetisaurusTrainerMixin, G2PTrainer, TopLevelMfaWorker
):
    """
    Top level trainer class for Phonetisaurus-style models
    """

    def __init__(
        self,
        **kwargs,
    ):
        self._data_source = os.path.splitext(os.path.basename(kwargs["dictionary_path"]))[0]
        super().__init__(**kwargs)
        self.ler = None
        self.wer = None

    @property
    def data_directory(self) -> str:
        """Data directory for trainer"""
        return self.working_directory

    @property
    def workflow_identifier(self) -> str:
        """Identifier for Phonetisaurus G2P trainer"""
        return "phonetisaurus_train_g2p"

    @property
    def configuration(self) -> MetaDict:
        """Configuration for G2P trainer"""
        config = super().configuration
        config.update({"dictionary_path": self.dictionary_model.path})
        return config

    def setup(self) -> None:
        """Setup for G2P training"""
        if self.initialized:
            return
        self.initialize_database()
        self.dictionary_setup()
        os.makedirs(self.phones_dir, exist_ok=True)
        self.initialize_training()
        self.initialized = True

    def train(self) -> None:
        """
        Train a G2P model
        """
        os.makedirs(self.working_log_directory, exist_ok=True)
        begin = time.time()
        self.train_alignments()
        self.log_debug(
            f"Aligning {len(self.g2p_training_dictionary)} words took {time.time() - begin} seconds"
        )
        begin = time.time()
        self.train_ngram_model()
        self.log_debug(
            f"Generating model for {len(self.g2p_training_dictionary)} words took {time.time() - begin} seconds"
        )
        self.finalize_training()

    def finalize_training(self) -> None:
        """Finalize training and run evaluation if specified"""
        if self.evaluation_mode:
            self.evaluate_g2p_model()

    @property
    def meta(self) -> MetaDict:
        """Metadata for exported G2P model"""
        from datetime import datetime

        from ..utils import get_mfa_version

        m = {
            "version": get_mfa_version(),
            "architecture": self.architecture,
            "train_date": str(datetime.now()),
            "phones": sorted(self.non_silence_phones),
            "graphemes": self.g2p_training_graphemes,
            "grapheme_order": self.grapheme_order,
            "phone_order": self.phone_order,
            "seq_sep": self.seq_sep,
            "evaluation": {},
            "training": {
                "num_words": len(self.g2p_training_dictionary),
                "num_graphemes": len(self.g2p_training_graphemes),
                "num_phones": len(self.non_silence_phones),
            },
        }

        if self.evaluation_mode:
            m["evaluation"]["num_words"] = len(self.g2p_validation_dictionary)
            m["evaluation"]["word_error_rate"] = self.wer
            m["evaluation"]["phone_error_rate"] = self.ler
        return m

    def evaluate_g2p_model(self) -> None:
        """
        Validate the G2P model against held out data
        """
        temp_model_path = os.path.join(self.working_log_directory, "g2p_model.zip")
        self.export_model(temp_model_path)
        temp_dir = os.path.join(self.working_directory, "validation")
        gen = PyniniValidator(
            g2p_model_path=temp_model_path,
            word_list=list(self.g2p_validation_dictionary.keys()),
            temporary_directory=temp_dir,
            num_jobs=self.num_jobs,
            num_pronunciations=self.num_pronunciations,
        )
        output = gen.generate_pronunciations()
        with open(os.path.join(temp_dir, "validation_output.txt"), "w", encoding="utf8") as f:
            for (orthography, pronunciations) in output.items():
                if not pronunciations:
                    continue
                for p in pronunciations:
                    if not p:
                        continue
                    f.write(f"{orthography}\t{p}\n")
        self.compute_validation_errors(output)

    def initialize_training(self) -> None:
        """Initialize training G2P model"""
        with self.session() as session:
            self.g2p_training_dictionary = {}
            pronunciations = (
                session.query(Word.word, Pronunciation.pronunciation)
                .join(Pronunciation.word)
                .filter(Word.word_type.in_([WordType.speech, WordType.clitic]))
            )
            for w, p in pronunciations:
                if w not in self.g2p_training_dictionary:
                    self.g2p_training_dictionary[w] = set()
                self.g2p_training_dictionary[w].add(p)
            if self.evaluation_mode:
                word_dict = self.g2p_training_dictionary
                words = sorted(word_dict.keys())
                total_items = len(words)
                validation_items = int(total_items * self.validation_proportion)
                validation_words = set(random.sample(words, validation_items))
                self.g2p_training_dictionary = {
                    k: v for k, v in word_dict.items() if k not in validation_words
                }
                self.g2p_validation_dictionary = {
                    k: v for k, v in word_dict.items() if k in validation_words
                }
                if self.debug:
                    with open(
                        os.path.join(self.working_directory, "validation_set.txt"),
                        "w",
                        encoding="utf8",
                    ) as f:
                        for word in self.g2p_validation_dictionary:
                            f.write(word + "\n")
            grapheme_count = 0
            phone_count = 0
            self.character_sets = set()
            for word, pronunciations in self.g2p_training_dictionary.items():
                # if re.match(r"\W", word) is not None:
                #    continue
                word = list(word)
                grapheme_count += len(word)
                self.g2p_training_graphemes.update(word)
                for p in pronunciations:
                    self.g2p_training_phones.update(p.split())
                    phone_count += len(p.split())
            self.log_debug(f"Graphemes in training data: {sorted(self.g2p_training_graphemes)}")
            self.log_debug(f"Phones in training data: {sorted(self.g2p_training_phones)}")
            self.log_debug(f"Averages phones per grapheme: {phone_count / grapheme_count}")

            if self.evaluation_mode:
                for word, pronunciations in self.g2p_validation_dictionary.items():
                    self.g2p_validation_graphemes.update(word)
                    for p in pronunciations:
                        self.g2p_validation_phones.update(p.split())
                self.log_debug(
                    f"Graphemes in validation data: {sorted(self.g2p_validation_graphemes)}"
                )
                self.log_debug(f"Phones in validation data: {sorted(self.g2p_validation_phones)}")
                grapheme_diff = sorted(self.g2p_validation_graphemes - self.g2p_training_graphemes)
                phone_diff = sorted(self.g2p_validation_phones - self.g2p_training_phones)
                if grapheme_diff:
                    self.log_warning(
                        f"The following graphemes appear only in the validation set: {', '.join(grapheme_diff)}"
                    )
                if phone_diff:
                    self.log_warning(
                        f"The following phones appear only in the validation set: {', '.join(phone_diff)}"
                    )



"""Class definitions for aligning with pretrained acoustic models"""
from __future__ import annotations

import os
import shutil
import time
from typing import TYPE_CHECKING, List, Optional

from sqlalchemy.orm import Session

from montreal_forced_aligner.abc import TopLevelMfaWorker
from montreal_forced_aligner.alignment.base import CorpusAligner
from montreal_forced_aligner.data import PhoneType
from montreal_forced_aligner.db import (
    Dictionary,
    Grapheme,
    Phone,
    PhoneInterval,
    Speaker,
    Utterance,
    WordInterval,
)
from montreal_forced_aligner.exceptions import AlignerError, KaldiProcessingError
from montreal_forced_aligner.helper import load_configuration, parse_old_features
from montreal_forced_aligner.models import AcousticModel
from montreal_forced_aligner.online.alignment import (
    OnlineAlignmentArguments,
    OnlineAlignmentFunction,
)
from montreal_forced_aligner.utils import log_kaldi_errors

if TYPE_CHECKING:
    from argparse import Namespace

    from montreal_forced_aligner.abc import MetaDict

__all__ = ["PretrainedAligner"]


class PretrainedAligner(CorpusAligner, TopLevelMfaWorker):
    """
    Class for aligning a dataset using a pretrained acoustic model

    Parameters
    ----------
    acoustic_model_path : str
        Path to acoustic model

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.base.CorpusAligner`
        For dictionary and corpus parsing parameters and alignment parameters
    :class:`~montreal_forced_aligner.abc.TopLevelMfaWorker`
        For top-level parameters
    """

    def __init__(
        self,
        acoustic_model_path: str,
        **kwargs,
    ):
        self.acoustic_model = AcousticModel(acoustic_model_path)
        kw = self.acoustic_model.parameters
        kw.update(kwargs)
        super().__init__(**kw)

    @property
    def working_directory(self) -> str:
        """Working directory"""
        return self.workflow_directory

    def setup_acoustic_model(self) -> None:
        """Set up the acoustic model"""
        self.acoustic_model.export_model(self.working_directory)
        os.makedirs(self.phones_dir, exist_ok=True)
        exist_check = os.path.exists(self.db_path)
        if not exist_check:
            self.initialize_database()
        for f in ["phones.txt", "graphemes.txt"]:
            path = os.path.join(self.working_directory, f)
            if os.path.exists(path):
                os.rename(path, os.path.join(self.phones_dir, f))
        dict_info = self.acoustic_model.meta.get("dictionaries", None)
        if not dict_info:
            return
        os.makedirs(self.dictionary_output_directory, exist_ok=True)
        self.oov_word = dict_info["oov_word"]
        self.silence_word = dict_info["silence_word"]
        self.bracketed_word = dict_info["bracketed_word"]
        self.use_g2p = dict_info["use_g2p"]
        self.laughter_word = dict_info["laughter_word"]
        self.clitic_marker = dict_info["clitic_marker"]
        self.position_dependent_phones = dict_info["position_dependent_phones"]
        self.compile_regexes()
        if not self.use_g2p:
            return
        dictionary_id_cache = {}
        with self.session() as session:
            for speaker_id, speaker_name, dictionary_id, dict_name, path in (
                session.query(
                    Speaker.id, Speaker.name, Dictionary.id, Dictionary.name, Dictionary.path
                )
                .outerjoin(Speaker.dictionary)
                .filter(Dictionary.default == False)  # noqa
            ):
                if speaker_id is not None:
                    self._speaker_ids[speaker_name] = speaker_id
                dictionary_id_cache[path] = dictionary_id
                self.dictionary_lookup[dict_name] = dictionary_id
            dictionary = (
                session.query(Dictionary).filter(Dictionary.default == True).first()  # noqa
            )
            if dictionary:
                self._default_dictionary_id = dictionary.id
                dictionary_id_cache[dictionary.path] = self._default_dictionary_id
                self.dictionary_lookup[dictionary.name] = dictionary.id
            for dict_name in dict_info["names"]:
                dictionary = Dictionary(
                    name=dict_name,
                    path=dict_name,
                    phone_set_type=self.phone_set_type,
                    root_temp_directory=self.dictionary_output_directory,
                    position_dependent_phones=self.position_dependent_phones,
                    clitic_marker=self.clitic_marker,
                    bracket_regex=self.bracket_regex.pattern,
                    clitic_cleanup_regex=self.clitic_cleanup_regex.pattern,
                    laughter_regex=self.laughter_regex.pattern,
                    default=dict_name == dict_info["default"],
                    use_g2p=self.use_g2p,
                    max_disambiguation_symbol=0,
                    silence_word=self.silence_word,
                    oov_word=self.oov_word,
                    bracketed_word=self.bracketed_word,
                    laughter_word=self.laughter_word,
                    optional_silence_phone=self.optional_silence_phone,
                )
                session.add(dictionary)
                session.flush()
                dictionary_id_cache[dict_name] = dictionary.id
                if dictionary.default:
                    self._default_dictionary_id = dictionary.id
                fst_path = os.path.join(self.acoustic_model.dirname, dict_name + ".fst")
                if os.path.exists(fst_path):
                    os.makedirs(dictionary.temp_directory, exist_ok=True)
                    shutil.copyfile(fst_path, os.path.join(dictionary.temp_directory, "L.fst"))
            phone_objs = []
            with open(self.phone_symbol_table_path, "r", encoding="utf8") as f:
                for line in f:
                    line = line.strip()
                    phone, mapping_id = line.split()
                    mapping_id = int(mapping_id)
                    phone_type = PhoneType.non_silence
                    if phone.startswith("#"):
                        phone_type = PhoneType.disambiguation
                    elif phone in self.kaldi_silence_phones:
                        phone_type = PhoneType.silence
                    phone_objs.append(
                        {
                            "id": mapping_id + 1,
                            "mapping_id": mapping_id,
                            "phone": phone,
                            "phone_type": phone_type,
                        }
                    )
            grapheme_objs = []
            with open(self.grapheme_symbol_table_path, "r", encoding="utf8") as f:
                for line in f:
                    line = line.strip()
                    grapheme, mapping_id = line.split()
                    mapping_id = int(mapping_id)
                    grapheme_objs.append(
                        {"id": mapping_id + 1, "mapping_id": mapping_id, "grapheme": grapheme}
                    )
            session.bulk_insert_mappings(Grapheme, grapheme_objs)
            session.bulk_insert_mappings(Phone, phone_objs)
            session.commit()

    def setup(self) -> None:
        """Setup for alignment"""
        if self.initialized:
            return
        begin = time.time()
        try:
            os.makedirs(self.working_log_directory, exist_ok=True)
            check = self.check_previous_run()
            if check:
                self.log_debug(
                    "There were some differences in the current run compared to the last one. "
                    "This may cause issues, run with --clean, if you hit an error."
                )
            self.setup_acoustic_model()
            self.load_corpus()
            if self.excluded_pronunciation_count:
                self.log_warning(
                    f"There were {self.excluded_pronunciation_count} pronunciations in the dictionary that "
                    f"were ignored for containing one of {len(self.excluded_phones)} phones not present in the"
                    f"trained acoustic model.  Please run `mfa validate` to get more details."
                )
            self.acoustic_model.validate(self)
            import logging

            logger = logging.getLogger(self.identifier)
            self.acoustic_model.log_details(logger)

        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise
        self.initialized = True
        self.log_debug(f"Setup for alignment in {time.time() - begin} seconds")

    @classmethod
    def parse_parameters(
        cls,
        config_path: Optional[str] = None,
        args: Optional[Namespace] = None,
        unknown_args: Optional[List[str]] = None,
    ) -> MetaDict:
        """
        Parse parameters from a config path or command-line arguments

        Parameters
        ----------
        config_path: str
            Config path
        args: :class:`~argparse.Namespace`
            Command-line arguments from argparse
        unknown_args: list[str], optional
            Extra command-line arguments

        Returns
        -------
        dict[str, Any]
            Configuration parameters
        """
        global_params = {}
        if config_path and os.path.exists(config_path):
            data = load_configuration(config_path)
            data = parse_old_features(data)
            for k, v in data.items():
                if k == "features":
                    global_params.update(v)
                else:
                    if v is None and k in cls.nullable_fields:
                        v = []
                    global_params[k] = v
        global_params.update(cls.parse_args(args, unknown_args))
        return global_params

    @property
    def configuration(self) -> MetaDict:
        """Configuration for aligner"""
        config = super().configuration
        config.update(
            {
                "acoustic_model": self.acoustic_model.name,
            }
        )
        return config

    @property
    def workflow_identifier(self) -> str:
        """Aligner identifier"""
        return "pretrained_aligner"

    def align_one_utterance(self, utterance: Utterance, session: Session) -> None:
        """
        Align a single utterance

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.db.Utterance`
            Utterance object to align
        session: :class:`~sqlalchemy.orm.session.Session`
            Session to use
        """
        dictionary = utterance.speaker.dictionary
        self.acoustic_model.export_model(self.working_directory)
        sox_string = utterance.file.sound_file.sox_string
        if not sox_string:
            sox_string = utterance.file.sound_file.sound_file_path
        text_int_path = os.path.join(self.working_directory, "text.int")
        with open(text_int_path, "w", encoding="utf8") as f:
            f.write(f"{utterance.kaldi_id} {utterance.normalized_text_int}\n")
        if utterance.features:
            feats_path = os.path.join(self.working_directory, "feats.scp")
            with open(feats_path, "w", encoding="utf8") as f:
                f.write(f"{utterance.kaldi_id} {utterance.features}\n")
        else:
            wav_path = os.path.join(self.working_directory, "wav.scp")
            segment_path = os.path.join(self.working_directory, "segments.scp")
            with open(wav_path, "w", encoding="utf8") as f:
                f.write(f"{utterance.file_id} {sox_string}\n")
            with open(segment_path, "w", encoding="utf8") as f:
                f.write(
                    f"{utterance.kaldi_id} {utterance.file_id} {utterance.begin} {utterance.end} {utterance.channel}\n"
                )
        if utterance.speaker.cmvn:
            cmvn_path = os.path.join(self.working_directory, "cmvn.scp")
            with open(cmvn_path, "w", encoding="utf8") as f:
                f.write(f"{utterance.speaker.id} {utterance.speaker.cmvn}\n")
        spk2utt_path = os.path.join(self.working_directory, "spk2utt.scp")
        utt2spk_path = os.path.join(self.working_directory, "utt2spk.scp")
        with open(spk2utt_path, "w") as f:
            f.write(f"{utterance.speaker.id} {utterance.kaldi_id}\n")
        with open(utt2spk_path, "w") as f:
            f.write(f"{utterance.kaldi_id} {utterance.speaker.id}\n")

        args = OnlineAlignmentArguments(
            0,
            self.db_path,
            os.path.join(self.working_directory, "align.log"),
            self.working_directory,
            sox_string,
            utterance.to_data(),
            self.mfcc_options,
            self.pitch_options,
            self.feature_options,
            self.align_options,
            self.alignment_model_path,
            self.tree_path,
            self.disambiguation_symbols_int_path,
            dictionary.lexicon_fst_path,
            dictionary.word_boundary_int_path,
            self.reversed_phone_mapping,
            self.optional_silence_phone,
            {self.silence_word},
        )
        func = OnlineAlignmentFunction(args)
        word_intervals, phone_intervals, log_likelihood = func.run()
        session.query(PhoneInterval).filter(PhoneInterval.utterance_id == utterance.id).delete()
        session.query(WordInterval).filter(WordInterval.utterance_id == utterance.id).delete()
        session.flush()
        for wi in word_intervals:
            session.add(WordInterval.from_ctm(wi, utterance))
        for pi in phone_intervals:
            session.add(PhoneInterval.from_ctm(pi, utterance))
        utterance.alignment_log_likelihood = log_likelihood
        session.commit()

    def align(self) -> None:
        """Run the aligner"""
        self.setup()
        done_path = os.path.join(self.working_directory, "done")
        dirty_path = os.path.join(self.working_directory, "dirty")
        if os.path.exists(done_path):
            self.log_info("Alignment already done, skipping.")
            return
        try:
            log_dir = os.path.join(self.working_directory, "log")
            os.makedirs(log_dir, exist_ok=True)
            self.compile_train_graphs()

            self.log_info("Performing first-pass alignment...")
            self.speaker_independent = True
            self.align_utterances()
            self.compile_information()
            if self.uses_speaker_adaptation:
                if self.alignment_model_path.endswith(".mdl"):
                    if os.path.exists(self.alignment_model_path.replace(".mdl", ".alimdl")):
                        raise AlignerError(
                            "Not using speaker independent model when it is available"
                        )
                self.calc_fmllr()

                self.speaker_independent = False
                assert self.alignment_model_path.endswith(".mdl")
                self.log_info("Performing second-pass alignment...")
                self.align_utterances()

                self.compile_information()
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise
        with open(done_path, "w"):
            pass


class DictionaryTrainer(PretrainedAligner):
    """
    Aligner for calculating pronunciation probabilities of dictionary entries

    Parameters
    ----------
    calculate_silence_probs: bool
        Flag for whether to calculate silence probabilities, default is False
    min_count: int
        Specifies the minimum count of words to include in derived probabilities,
        affects probabilities of infrequent words more, default is 1

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.pretrained.PretrainedAligner`
        For dictionary and corpus parsing parameters and alignment parameters
    """

    def __init__(
        self,
        calculate_silence_probs: bool = False,
        min_count: int = 1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.calculate_silence_probs = calculate_silence_probs
        self.min_count = min_count

    def export_lexicons(
        self, output_directory: str, silence_probabilities: Optional[bool] = False
    ) -> None:
        """
        Generate pronunciation probabilities for the dictionary

        Parameters
        ----------
        output_directory: str
            Directory in which to save new dictionaries
        silence_probabilities: bool
            Flag for whether to save silence probabilities as well

        See Also
        --------
        :func:`~montreal_forced_aligner.alignment.multiprocessing.GeneratePronunciationsFunction`
            Multiprocessing helper function for each job
        :meth:`.CorpusAligner.generate_pronunciations_arguments`
            Job method for generating arguments for helper function

        """
        self.compute_pronunciation_probabilities()
        os.makedirs(output_directory, exist_ok=True)
        with self.session() as session:
            for dictionary in session.query(Dictionary):
                self.export_lexicon(
                    dictionary.id,
                    os.path.join(output_directory, dictionary.name + ".dict"),
                    probability=True,
                    silence_probabilities=silence_probabilities,
                )



"""Validation classes"""

from montreal_forced_aligner.validation.corpus_validator import (
    PretrainedValidator,
    TrainingValidator,
    ValidationMixin,
)
from montreal_forced_aligner.validation.dictionary_validator import DictionaryValidator

__all__ = ["PretrainedValidator", "TrainingValidator", "ValidationMixin", "DictionaryValidator"]



"""Class definitions for training G2P models"""
from __future__ import annotations

import itertools
import multiprocessing as mp
import operator
import os
import queue
import random
import re
import shutil
import subprocess
import time
from typing import Any, List, NamedTuple, Optional, Set

import tqdm

from montreal_forced_aligner.abc import MetaDict, MfaWorker, TopLevelMfaWorker, TrainerMixin
from montreal_forced_aligner.data import WordType
from montreal_forced_aligner.db import Pronunciation, Word
from montreal_forced_aligner.dictionary.multispeaker import MultispeakerDictionaryMixin
from montreal_forced_aligner.exceptions import KaldiProcessingError, PyniniAlignmentError
from montreal_forced_aligner.g2p.generator import PyniniValidator
from montreal_forced_aligner.models import G2PModel
from montreal_forced_aligner.utils import Stopped, thirdparty_binary

try:
    import pynini
    import pywrapfst
    from pynini import Fst, TokenType
    from pynini.lib import rewrite
    from pywrapfst import convert

    G2P_DISABLED = False

except ImportError:
    pynini = None
    pywrapfst = None
    TokenType = Optional[str]
    rewrite = None
    Fst = None

    def convert(x):
        """stub function"""
        pass

    G2P_DISABLED = True


Labels = List[Any]

TOKEN_TYPES = ["byte", "utf8"]
INF = float("inf")
RAND_MAX = 32767

__all__ = ["RandomStartWorker", "PyniniTrainer", "G2PTrainer"]


class RandomStart(NamedTuple):
    """Parameters for random starts"""

    idx: int
    seed: int
    input_far_path: str
    output_far_path: str
    cg_path: str
    tempdir: str
    train_opts: List[str]


def _get_far_labels(far_path: str) -> Set[int]:
    """Extracts label set from acceptors in a FAR.
    Args:
      far_path: path to FAR file.
    Returns:
      A set of integer labels found in the FAR.
    """
    labels: Set[int] = set()
    reader = pywrapfst.FarReader.open(far_path)
    while not reader.done():
        fst = reader.get_fst()
        assert fst.properties(pywrapfst.ACCEPTOR, True) == pywrapfst.ACCEPTOR
        for state in fst.states():
            labels.update(arc.ilabel for arc in fst.arcs(state))
        next(reader)
    assert not reader.error()
    return labels


class RandomStartWorker(mp.Process):
    """
    Random start worker
    """

    def __init__(
        self,
        job_name: int,
        job_q: mp.Queue,
        return_queue: mp.Queue,
        log_file: str,
        stopped: Stopped,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.job_q = job_q
        self.return_queue = return_queue
        self.log_file = log_file
        self.stopped = stopped
        self.finished = Stopped()

    def run(self) -> None:
        """Run the random start worker"""
        with open(self.log_file, "w", encoding="utf8") as log_file:
            while True:
                try:
                    args = self.job_q.get(timeout=1)
                except queue.Empty:
                    break
                if self.stopped.stop_check():
                    continue
                try:
                    start = time.time()
                    # Randomize channel model.
                    rfst_path = os.path.join(args.tempdir, f"random-{args.seed:05d}.fst")
                    afst_path = os.path.join(args.tempdir, f"aligner-{args.seed:05d}.fst")
                    likelihood_path = afst_path.replace(".fst", ".like")
                    if not os.path.exists(afst_path):
                        cmd = [
                            thirdparty_binary("baumwelchrandomize"),
                            f"--seed={args.seed}",
                            args.cg_path,
                            rfst_path,
                        ]
                        subprocess.check_call(cmd, stderr=log_file, env=os.environ)
                        random_end = time.time()
                        log_file.write(
                            f"{args.seed} randomization took {random_end - start} seconds\n"
                        )
                        # Train on randomized channel model.

                        likelihood = INF
                        cmd = [
                            thirdparty_binary("baumwelchtrain"),
                            *args.train_opts,
                            args.input_far_path,
                            args.output_far_path,
                            rfst_path,
                            afst_path,
                        ]
                        log_file.write(f"{args.seed} train command: {' '.join(cmd)}\n")
                        log_file.flush()
                        with subprocess.Popen(
                            cmd, stderr=subprocess.PIPE, text=True, env=os.environ
                        ) as proc:
                            # Parses STDERR to capture the likelihood.
                            for line in proc.stderr:  # type: ignore
                                log_file.write(line)
                                log_file.flush()
                                line = line.rstrip()
                                match = re.match(r"INFO: Iteration \d+: (-?\d*(\.\d*)?)", line)
                                assert match, line
                                likelihood = float(match.group(1))
                                self.return_queue.put(1)
                            with open(likelihood_path, "w") as f:
                                f.write(str(likelihood))
                        log_file.write(
                            f"{args.seed} training took {time.time() - random_end} seconds\n"
                        )
                    else:
                        with open(likelihood_path, "r") as f:
                            likelihood = f.read().strip()
                    self.return_queue.put((afst_path, likelihood))
                except Exception:
                    self.stopped.stop()
                    e = KaldiProcessingError([self.log_file])
                    e.job_name = self.job_name
                    self.return_queue.put(e)
        self.finished.stop()
        return


class G2PTrainer(MfaWorker, TrainerMixin):
    """
    Abstract mixin class for G2P training

    Parameters
    ----------
    validation_proportion: float
        Proportion of words to use as the validation set, defaults to 0.1, only used if ``evaluate`` is True
    num_pronunciations: int
        Number of pronunciations to generate
    evaluation_mode: bool
        Flag for whether to evaluate the model performance on an validation set

    See Also
    --------
    :class:`~montreal_forced_aligner.abc.MfaWorker`
        For base MFA parameters
    :class:`~montreal_forced_aligner.abc.TrainerMixin`
        For base trainer parameters

    Attributes
    ----------
    g2p_training_dictionary: dict[str, list[str]]
        Dictionary of words to pronunciations to train from
    g2p_validation_dictionary: dict[str, list[str]]
        Dictionary of words to pronunciations to validate performance against
    g2p_graphemes: set[str]
        Set of graphemes in the training set
    """

    def __init__(
        self,
        validation_proportion: float = 0.1,
        num_pronunciations: int = 0,
        evaluation_mode: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.evaluation_mode = evaluation_mode
        self.validation_proportion = validation_proportion
        self.num_pronunciations = num_pronunciations
        self.g2p_training_dictionary = {}
        self.g2p_validation_dictionary = None
        self.g2p_training_graphemes = set()
        self.g2p_validation_graphemes = set()
        self.g2p_training_phones = set()
        self.g2p_validation_phones = set()


class PyniniTrainerMixin:
    """
    Mixin for training Pynini G2P models

    Parameters
    ----------
    order: int
        Order of the ngram model, defaults to 7
    random_starts: int
        Number of random starts to use in initialization, defaults to 25
    seed: int
        Seed for randomization, defaults to 1917
    delta: float
        Comparison/quantization delta for Baum-Welch training, defaults to 1/1024
    alpha: float
        Step size reduction power parameter for Baum-Welch training;
        full standard batch EM is run (not stepwise) if set to 0, defaults to 1.0
    batch_size:int
        Batch size for Baum-Welch training, defaults to 200
    num_iterations:int
        Maximum number of iterations to use in Baum-Welch training, defaults to 10
    smoothing_method:str
        Smoothing method for the ngram model, defaults to "kneser_ney"
    pruning_method:str
        Pruning method for pruning the ngram model, defaults to "relative_entropy"
    model_size: int
        Target number of ngrams for pruning, defaults to 1000000
    insertions: bool
        Flag for whether to allow for insertions, default True
    deletions: bool
        Flag for whether to allow for deletions, default True
    fst_default_cache_gc: str
        String to pass to OpenFst binaries for GC behavior
    fst_default_cache_gc_limit: str
        String to pass to OpenFst binaries for GC behavior
    """

    def __init__(
        self,
        order: int = 8,
        random_starts: int = 25,
        seed: int = 1917,
        delta: float = 1 / 1024,
        alpha: float = 1.0,
        batch_size: int = 800,
        num_iterations: int = 10,
        smoothing_method: str = "kneser_ney",
        pruning_method: str = "relative_entropy",
        model_size: int = 1000000,
        insertions: bool = True,
        deletions: bool = True,
        fst_default_cache_gc="",
        fst_default_cache_gc_limit="",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not hasattr(self, "_data_source"):
            self._data_source = None
        self.order = order
        self.random_starts = random_starts
        self.seed = seed
        self.delta = delta
        self.alpha = alpha
        self.batch_size = batch_size
        self.num_iterations = num_iterations
        self.smoothing_method = smoothing_method
        self.pruning_method = pruning_method
        self.model_size = model_size
        self.insertions = insertions
        self.deletions = deletions
        self.fst_default_cache_gc = fst_default_cache_gc
        self.fst_default_cache_gc_limit = fst_default_cache_gc_limit
        self._sym_path = None
        self._fst_path = None
        self.input_sym_path = None
        self.input_token_type = "utf8"
        self.output_token_type = "utf8"

    @property
    def data_source_identifier(self) -> str:
        """Dictionary name"""
        return self._data_source

    def train_iteration(self) -> None:
        """Train iteration, not used"""
        pass

    @property
    def architecture(self) -> str:
        """Pynini"""
        return "pynini"

    @property
    def input_far_path(self) -> str:
        """Path to store grapheme archive"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.g.far")

    @property
    def output_far_path(self) -> str:
        """Path to store phone archive"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.p.far")

    @property
    def cg_path(self) -> str:
        """Path to covering grammar FST"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.cg.fst")

    @property
    def align_path(self) -> str:
        """Path to store alignment models"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.align.fst")

    @property
    def afst_path(self) -> str:
        """Path to store aligned FSTs"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.afst.far")

    @property
    def input_path(self) -> str:
        """Path to temporary file to store grapheme training data"""
        return os.path.join(self.working_directory, f"input_{self.data_source_identifier}.txt")

    @property
    def output_path(self) -> str:
        """Path to temporary file to store phone training data"""
        return os.path.join(self.working_directory, f"output_{self.data_source_identifier}.txt")

    def generate_model(self) -> None:
        """
        Generate an ngram G2P model from FAR strings
        """
        assert os.path.exists(self.far_path)
        if os.path.exists(self.fst_path):
            self.log_info("Model building already done, skipping!")
            return
        with open(
            os.path.join(self.working_log_directory, "model.log"), "w", encoding="utf8"
        ) as logf:
            ngramcount_proc = subprocess.Popen(
                [
                    thirdparty_binary("ngramcount"),
                    "--require_symbols=false",
                    f"--order={self.order}",
                    self.far_path,
                ],
                stderr=logf,
                stdout=subprocess.PIPE,
                env=os.environ,
            )

            ngrammake_proc = subprocess.Popen(
                [thirdparty_binary("ngrammake"), f"--method={self.smoothing_method}"],
                stdin=ngramcount_proc.stdout,
                stderr=logf,
                stdout=subprocess.PIPE,
                env=os.environ,
            )

            ngramshrink_proc = subprocess.Popen(
                [
                    thirdparty_binary("ngramshrink"),
                    f"--method={self.pruning_method}",
                    f"--target_number_of_ngrams={self.model_size}",
                ],
                stdin=ngrammake_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=logf,
                env=os.environ,
            )

            fstencode_proc = subprocess.Popen(
                [thirdparty_binary("fstencode"), "--decode", "-", self.encoder_path, "-"],
                stdin=ngramshrink_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=logf,
                env=os.environ,
            )
            sort_proc = subprocess.Popen(
                [
                    thirdparty_binary("fstarcsort"),
                    "-",
                    self.fst_path,
                ],
                stdin=fstencode_proc.stdout,
                stderr=logf,
                env=os.environ,
            )
            sort_proc.communicate()

    @property
    def fst_path(self):
        """Internal temporary FST file"""
        if self._fst_path is not None:
            return self._fst_path
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.fst")

    @property
    def far_path(self):
        """Internal temporary FAR file"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.far")

    @property
    def encoder_path(self):
        """Internal temporary encoder file"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.enc")

    @property
    def sym_path(self):
        """Internal temporary symbol file"""
        if self._sym_path is not None:
            return self._sym_path
        return os.path.join(self.working_directory, "phones.txt")

    def align_g2p(self):
        """Runs the entire alignment regimen."""
        self._lexicon_covering()
        self._alignments()
        self._encode()

    @staticmethod
    def _narcs(f: Fst) -> int:
        """Computes the number of arcs in an FST."""
        return sum(f.num_arcs(state) for state in f.states())

    def _lexicon_covering(
        self,
    ) -> None:
        """Builds covering grammar and lexicon FARs."""
        # Sets of labels for the covering grammar.
        with open(
            os.path.join(self.working_log_directory, "covering_grammar.log"), "w", encoding="utf8"
        ) as log_file:
            com = [
                thirdparty_binary("farcompilestrings"),
                "--fst_type=compact",
            ]
            if self.input_token_type != "utf8":
                com.append("--token_type=symbol")
                com.append(
                    f"--symbols={self.input_token_type}",
                )
                com.append("--unknown_symbol=<unk>")
            else:
                com.append("--token_type=utf8")
            com.extend([self.input_path, self.input_far_path])
            print(" ".join(com), file=log_file)
            subprocess.check_call(com, env=os.environ, stderr=log_file, stdout=log_file)
            com = [
                thirdparty_binary("farcompilestrings"),
                "--fst_type=compact",
                "--token_type=symbol",
                f"--symbols={self.phone_symbol_table_path}",
                self.output_path,
                self.output_far_path,
            ]
            print(" ".join(com), file=log_file)
            subprocess.check_call(com, env=os.environ, stderr=log_file, stdout=log_file)
            ilabels = _get_far_labels(self.input_far_path)
            print(ilabels, file=log_file)
            olabels = _get_far_labels(self.output_far_path)
            print(olabels, file=log_file)
            cg = pywrapfst.VectorFst()
            state = cg.add_state()
            cg.set_start(state)
            one = pywrapfst.Weight.one(cg.weight_type())
            for ilabel, olabel in itertools.product(ilabels, olabels):
                cg.add_arc(state, pywrapfst.Arc(ilabel, olabel, one, state))
            # Handles epsilons, carefully avoiding adding a useless 0:0 label.
            if self.insertions:
                for olabel in olabels:
                    cg.add_arc(state, pywrapfst.Arc(0, olabel, one, state))
            if self.deletions:
                for ilabel in ilabels:
                    cg.add_arc(state, pywrapfst.Arc(ilabel, 0, one, state))
            cg.set_final(state)
            assert cg.verify(), "Label acceptor is ill-formed"
            cg.write(self.cg_path)

    def _alignments(self) -> None:
        """Trains the aligner and constructs the alignments FAR."""
        if not os.path.exists(self.align_path):
            self.log_info("Training aligner")
            train_opts = []
            if self.batch_size:
                train_opts.append(f"--batch_size={self.batch_size}")
            if self.delta:
                train_opts.append(f"--delta={self.delta}")
            if self.fst_default_cache_gc:
                train_opts.append(f"--fst_default_cache_gc={self.fst_default_cache_gc}")
            if self.fst_default_cache_gc_limit:
                train_opts.append(
                    f"--fst_default_cache_gc_limit={self.fst_default_cache_gc_limit}"
                )
            if self.alpha:
                train_opts.append(f"--alpha={self.alpha}")
            if self.num_iterations:
                train_opts.append(f"--max_iters={self.num_iterations}")
            # Constructs the actual command vectors (plus an index for logging
            # purposes).
            random.seed(self.seed)
            starts = [
                (
                    RandomStart(
                        idx,
                        seed,
                        self.input_far_path,
                        self.output_far_path,
                        self.cg_path,
                        self.working_directory,
                        train_opts,
                    )
                )
                for (idx, seed) in enumerate(
                    random.sample(range(1, RAND_MAX), self.random_starts), 1
                )
            ]
            stopped = Stopped()
            num_commands = len(starts)
            job_queue = mp.JoinableQueue()
            fst_likelihoods = {}
            # Actually runs starts.
            self.log_info("Calculating alignments...")
            begin = time.time()
            with tqdm.tqdm(
                total=num_commands * self.num_iterations, disable=getattr(self, "quiet", False)
            ) as pbar:
                for start in starts:
                    job_queue.put(start)
                error_dict = {}
                return_queue = mp.Queue()
                procs = []
                for i in range(self.num_jobs):
                    log_path = os.path.join(self.working_log_directory, f"baumwelch.{i}.log")
                    p = RandomStartWorker(
                        i,
                        job_queue,
                        return_queue,
                        log_path,
                        stopped,
                    )
                    procs.append(p)
                    p.start()

                while True:
                    try:
                        result = return_queue.get(timeout=1)
                        if isinstance(result, Exception):
                            error_dict[getattr(result, "job_name", 0)] = result
                            continue
                        if stopped.stop_check():
                            continue
                    except queue.Empty:
                        for proc in procs:
                            if not proc.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    if isinstance(result, int):
                        pbar.update(result)
                    else:
                        fst_likelihoods[result[0]] = result[1]
                for p in procs:
                    p.join()
            if error_dict:
                raise PyniniAlignmentError(error_dict)
            (best_fst, best_likelihood) = min(fst_likelihoods.items(), key=operator.itemgetter(1))
            self.log_info(f"Best likelihood: {best_likelihood}")
            self.log_debug(
                f"Ran {self.random_starts} random starts in {time.time() - begin} seconds"
            )
            # Moves best likelihood solution to the requested location.
            shutil.move(best_fst, self.align_path)
        cmd = [thirdparty_binary("baumwelchdecode")]
        if self.fst_default_cache_gc:
            cmd.append(f"--fst_default_cache_gc={self.fst_default_cache_gc}")
        if self.fst_default_cache_gc_limit:
            cmd.append(f"--fst_default_cache_gc_limit={self.fst_default_cache_gc_limit}")
        cmd.append(self.input_far_path)
        cmd.append(self.output_far_path)
        cmd.append(self.align_path)
        cmd.append(self.afst_path)
        self.log_debug(f"Subprocess call: {cmd}")
        subprocess.check_call(cmd, env=os.environ)
        self.log_info("Completed computing alignments!")

    def _encode(self) -> None:
        """Encodes the alignments."""
        self.log_info("Encoding the alignments as FSAs")
        subprocess.check_call(
            [
                thirdparty_binary("farencode"),
                "--encode_labels",
                self.afst_path,
                self.encoder_path,
                self.far_path,
            ],
            env=os.environ,
        )
        self.log_info(f"Success! FAR path: {self.far_path}; encoder path: {self.encoder_path}")


class PyniniTrainer(
    MultispeakerDictionaryMixin, PyniniTrainerMixin, G2PTrainer, TopLevelMfaWorker
):
    """
    Top-level G2P trainer that uses Pynini functionality

    See Also
    --------
    :class:`~montreal_forced_aligner.g2p.trainer.G2PTrainer`
        For base G2P training parameters
    :class:`~montreal_forced_aligner.abc.TopLevelMfaWorker`
        For top-level parameters
    """

    def __init__(
        self,
        **kwargs,
    ):
        self._data_source = os.path.splitext(os.path.basename(kwargs["dictionary_path"]))[0]
        super().__init__(**kwargs)
        self._fst_path = None
        self._sym_path = None
        self.position_dependent_phones = False
        self.wer = None
        self.ler = None

    @property
    def data_directory(self) -> str:
        """Data directory for trainer"""
        return self.working_directory

    @property
    def workflow_identifier(self) -> str:
        """Identifier for Pynini G2P trainer"""
        return "pynini_train_g2p"

    @property
    def configuration(self) -> MetaDict:
        """Configuration for G2P trainer"""
        config = super().configuration
        config.update({"dictionary_path": self.dictionary_model.path})
        return config

    def setup(self) -> None:
        """Setup for G2P training"""
        if self.initialized:
            return
        self.initialize_database()
        self.dictionary_setup()
        os.makedirs(self.phones_dir, exist_ok=True)
        self._write_phone_symbol_table()
        self._write_grapheme_symbol_table()
        os.makedirs(self.working_log_directory, exist_ok=True)
        self.initialize_training()
        self.initialized = True

    @property
    def meta(self) -> MetaDict:
        """Metadata for exported G2P model"""
        from datetime import datetime

        from ..utils import get_mfa_version

        m = {
            "version": get_mfa_version(),
            "architecture": self.architecture,
            "train_date": str(datetime.now()),
            "phones": sorted(self.non_silence_phones),
            "graphemes": self.g2p_training_graphemes,
            "evaluation": {},
            "training": {
                "num_words": len(self.g2p_training_dictionary),
                "num_graphemes": len(self.g2p_training_graphemes),
                "num_phones": len(self.non_silence_phones),
            },
        }

        if self.evaluation_mode:
            m["evaluation"]["num_words"] = len(self.g2p_validation_dictionary)
            m["evaluation"]["word_error_rate"] = self.wer
            m["evaluation"]["phone_error_rate"] = self.ler
        return m

    def initialize_training(self) -> None:
        """Initialize training G2P model"""
        random.seed(self.seed)
        self._sym_path = self.phone_symbol_table_path
        self.output_token_type = pynini.SymbolTable.read_text(self.phone_symbol_table_path)
        with self.session() as session:
            self.g2p_training_dictionary = {}
            pronunciations = (
                session.query(Word.word, Pronunciation.pronunciation)
                .join(Pronunciation.word)
                .filter(Word.word_type.in_([WordType.speech, WordType.clitic]))
            )
            for w, p in pronunciations:
                if w not in self.g2p_training_dictionary:
                    self.g2p_training_dictionary[w] = set()
                self.g2p_training_dictionary[w].add(p)
            if self.evaluation_mode:
                word_dict = self.g2p_training_dictionary
                words = sorted(word_dict.keys())
                total_items = len(words)
                validation_items = int(total_items * self.validation_proportion)
                validation_words = random.sample(words, validation_items)
                self.g2p_training_dictionary = {
                    k: v for k, v in word_dict.items() if k not in validation_words
                }
                self.g2p_validation_dictionary = {
                    k: v for k, v in word_dict.items() if k in validation_words
                }
                if self.debug:
                    with open(
                        os.path.join(self.working_directory, "validation_set.txt"),
                        "w",
                        encoding="utf8",
                    ) as f:
                        for word in self.g2p_validation_dictionary:
                            f.write(word + "\n")

            with open(self.input_path, "w", encoding="utf8") as inf, open(
                self.output_path, "w", encoding="utf8"
            ) as outf:
                for word, pronunciations in self.g2p_training_dictionary.items():
                    if re.match(r"\W", word) is not None:
                        continue
                    self.g2p_training_graphemes.update(word)
                    for p in pronunciations:
                        self.g2p_training_phones.update(p.split())
                        print(word, file=inf)
                        print(p, file=outf)
            self.log_debug(f"Graphemes in training data: {sorted(self.g2p_training_graphemes)}")
            self.log_debug(f"Phones in training data: {sorted(self.g2p_training_phones)}")
            if self.evaluation_mode:
                for word, pronunciations in self.g2p_validation_dictionary.items():
                    self.g2p_validation_graphemes.update(word)
                    for p in pronunciations:
                        self.g2p_validation_phones.update(p.split())
                self.log_debug(
                    f"Graphemes in validation data: {sorted(self.g2p_validation_graphemes)}"
                )
                self.log_debug(f"Phones in validation data: {sorted(self.g2p_validation_phones)}")
                grapheme_diff = sorted(self.g2p_validation_graphemes - self.g2p_training_graphemes)
                phone_diff = sorted(self.g2p_validation_phones - self.g2p_training_phones)
                if grapheme_diff:
                    self.log_warning(
                        f"The following graphemes appear only in the validation set: {', '.join(grapheme_diff)}"
                    )
                if phone_diff:
                    self.log_warning(
                        f"The following phones appear only in the validation set: {', '.join(phone_diff)}"
                    )

    def clean_up(self) -> None:
        """
        Clean up temporary files
        """
        if self.debug:
            return
        for name in os.listdir(self.working_directory):
            path = os.path.join(self.working_directory, name)
            if os.path.isdir(path):
                shutil.rmtree(path, ignore_errors=True)
            elif not name.endswith(".log"):
                os.remove(path)

    def export_model(self, output_model_path: str) -> None:
        """
        Export G2P model to specified path

        Parameters
        ----------
        output_model_path:str
            Path to export model
        """
        directory, filename = os.path.split(output_model_path)
        basename, _ = os.path.splitext(filename)
        models_temp_dir = os.path.join(self.working_directory, "model_archive_temp")
        model = G2PModel.empty(basename, root_directory=models_temp_dir)
        model.add_meta_file(self)
        model.add_fst_model(self.working_directory)
        model.add_sym_path(self.working_directory)
        if directory:
            os.makedirs(directory, exist_ok=True)
        basename, _ = os.path.splitext(output_model_path)
        model.dump(basename)
        model.clean_up()
        # self.clean_up()
        self.log_info(f"Saved model to {output_model_path}")

    def train(self) -> None:
        """
        Train a G2P model
        """
        os.makedirs(self.working_log_directory, exist_ok=True)
        begin = time.time()
        if os.path.exists(self.far_path) and os.path.exists(self.encoder_path):
            self.log_info("Alignment already done, skipping!")
        else:
            self.align_g2p()
            self.log_debug(
                f"Aligning {len(self.g2p_training_dictionary)} words took {time.time() - begin} seconds"
            )
        begin = time.time()
        self.generate_model()
        self.log_debug(
            f"Generating model for {len(self.g2p_training_dictionary)} words took {time.time() - begin} seconds"
        )
        self.finalize_training()

    def finalize_training(self) -> None:
        """Finalize training"""
        shutil.copyfile(self.fst_path, os.path.join(self.working_directory, "model.fst"))
        shutil.copyfile(
            self.phone_symbol_table_path, os.path.join(self.working_directory, "phones.txt")
        )
        if self.evaluation_mode:
            self.evaluate_g2p_model()

    def evaluate_g2p_model(self) -> None:
        """
        Validate the G2P model against held out data
        """
        temp_model_path = os.path.join(self.working_log_directory, "g2p_model.zip")
        self.export_model(temp_model_path)

        gen = PyniniValidator(
            g2p_model_path=temp_model_path,
            word_list=list(self.g2p_validation_dictionary.keys()),
            temporary_directory=os.path.join(self.working_directory, "validation"),
            num_jobs=self.num_jobs,
            num_pronunciations=self.num_pronunciations,
        )
        gen.evaluate_g2p_model(self.g2p_training_dictionary)
        self.wer = gen.wer
        self.ler = gen.ler



"""Classes for validating dictionaries"""
import os
import shutil
import typing

from montreal_forced_aligner.g2p.generator import PyniniValidator
from montreal_forced_aligner.g2p.trainer import PyniniTrainer


class DictionaryValidator(PyniniTrainer):
    """
    Mixin class for performing validation on a corpus

    Parameters
    ----------
    g2p_model_path: str, optional
        Path to pretrained G2P model
    g2p_threshold: float, optional
        Threshold for pruning pronunciations, defaults to 1.5, which returns the optimal pronunciations and those with scores less than 1.5 times
        the optimal pronunciation's score. Increase
        to allow for more suboptimal pronunciations

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.base.CorpusAligner`
        For corpus, dictionary, and alignment parameters

    Attributes
    ----------
    printer: TerminalPrinter
        Printer for output messages
    """

    def __init__(
        self,
        g2p_model_path: typing.Optional[str] = None,
        g2p_threshold: float = 1.5,
        **kwargs,
    ):
        kwargs["clean"] = True
        super().__init__(**kwargs)
        self.g2p_model_path = g2p_model_path
        self.g2p_threshold = g2p_threshold

    @property
    def workflow_identifier(self) -> str:
        """Identifier for validation"""
        return "validate_dictionary"

    def setup(self) -> None:
        """Set up the dictionary validator"""
        if self.initialized:
            return
        self.initialize_database()
        self.dictionary_setup()
        self.write_lexicon_information()
        if self.g2p_model_path is None:
            self.log_info("Not using a pretrained G2P model, training from the dictionary...")
            self.initialize_training()
            self.train()
            self.g2p_model_path = os.path.join(self.working_log_directory, "g2p_model.zip")
            self.export_model(self.g2p_model_path)
        else:
            self.initialize_training()
        self.initialized = True

    def validate(self, output_path: typing.Optional[str] = None) -> None:
        """
        Validate the dictionary

        Parameters
        ----------
        output_path: str, optional
            Path to save scored CSV
        """
        self.setup()

        gen = PyniniValidator(
            g2p_model_path=self.g2p_model_path,
            word_list=list(self.g2p_training_dictionary.keys()),
            temporary_directory=os.path.join(self.working_directory, "validation"),
            num_jobs=self.num_jobs,
            num_pronunciations=self.num_pronunciations,
        )
        gen.evaluate_g2p_model(self.g2p_training_dictionary)
        if output_path is not None:
            shutil.copyfile(gen.evaluation_csv_path, output_path)
            self.log_info(f"Wrote scores to {output_path}")
